import numpy as np
from zest import zest
from plaster.run.calib.calib import Calib, RegIllum, RegPSF, CalibIdentity
from plaster.tools.log.log import debug


def zest_reg_illum():
    def it_sets():
        reg_illum = RegIllum(2, 512, 5)

        vals0 = 100 * (0.5 + np.random.uniform(size=(5, 5)))
        reg_illum.set_ch(0, vals0)

        vals1 = 80 * (0.5 * np.random.uniform(size=(5, 5)))
        reg_illum.set_ch(1, vals1)

        assert np.min(reg_illum.reg_illum[0]) == 1.0
        assert np.min(reg_illum.reg_illum[1]) == 1.0
        assert not np.all(reg_illum.reg_illum[0] == 1.0)
        assert not np.all(reg_illum.reg_illum[1] == 1.0)
        assert not np.all(reg_illum.reg_illum[0] == reg_illum.reg_illum[1])

    def it_interpolates():
        reg_illum = RegIllum(1, 512, 5)
        vals = 100 * np.ones((5, 5))
        vals[0, 0] = 50
        reg_illum.set_ch(0, vals)
        im = reg_illum.interp(0)
        assert im.shape == (512, 512)
        assert np.allclose(im[0, 0], 2.0)  # Because it is half as bright its bal is 2.0
        assert np.allclose(im[511, 511], 1.0)

    zest()


def zest_reg_psf():
    def it_sample_params_grid_single_channel():
        reg_psf = RegPSF.fixture(n_channels=1)

        # Half the width of one corner
        reg_psf.params[0, 0, :, 0] *= 0.5
        samples = reg_psf.sample_params_grid(n_divs=10)
        assert samples.shape == (1, 10, 10, 3)

        # The corner parameter was changed from 1.8 to 0.9
        # so ensure that only that corner changed
        assert np.all(np.abs(samples[0, 4:, 0, 0] - 1.8) < 0.05)
        assert np.abs(samples[0, 0, 0, 0] - 0.9) < 0.05
        assert np.abs(samples[0, 1, 0, 0] - 1.0) < 0.05
        assert np.abs(samples[0, 2, 0, 0] - 1.6) < 0.05

    def it_sample_params_grid_multi_channel():
        reg_psf = RegPSF.fixture(n_channels=2)

        # Half the second channel width
        reg_psf.params[1, :, :, 0] *= 0.5
        samples = reg_psf.sample_params_grid(n_divs=10)
        assert samples.shape == (2, 10, 10, 3)

        # Check that the two channels have different values
        assert np.all(np.abs(samples[0, :, :, 0] - 1.8) < 0.05)
        assert np.all(np.abs(samples[1, :, :, 0] - 0.9) < 0.05)

    zest()


def zest_calib():
    def it_adds_reg_illum():
        reg_illum = RegIllum(2, 512, 5)
        calib = Calib()
        calib.add_reg_illum(reg_illum, CalibIdentity("foo"))
        test_path = "/tmp/_test.calib"
        calib.save_file(test_path)
        calib = Calib.load_file(test_path, CalibIdentity("foo"))
        assert np.all(calib.reg_illum().reg_illum == reg_illum.reg_illum)

    def it_adds_reg_psf():
        reg_psf = RegPSF.fixture(n_channels=2)
        calib = Calib()
        calib.add_reg_psf(reg_psf, CalibIdentity("foo"))
        test_path = "/tmp/_test.calib"
        calib.save_file(test_path)
        calib = Calib.load_file(test_path, CalibIdentity("foo"))
        assert np.all(calib.reg_psf().params == reg_psf.params)

    def adds_without_ident():
        reg_psf = RegPSF.fixture(n_channels=2)
        test_path = "/tmp/_test.calib"

        def it_enforces_identity():
            calib = Calib()
            calib.add_reg_psf(reg_psf)
            with zest.raises(ValueError, in_args="calib identity not specified"):
                calib.save_file(test_path)

        def it_rewrites_identity():
            calib = Calib()
            calib.add_reg_psf(reg_psf)
            calib.set_identity("me")
            calib.save_file(test_path)
            _calib = Calib.load_file(test_path, "me")
            assert np.all(_calib.reg_psf().params == reg_psf.params)

        zest()

    zest()
