import ctypes as c
import pathlib
from contextlib import contextmanager, redirect_stderr, redirect_stdout
from io import StringIO

import numpy as np
import pandas as pd
from plaster.run.error_model import GainModel
from plaster.run.nn_v2.c.build import build_dev
from plaster.run.nn_v2.nn_v2_result import NNV2Result
from plaster.run.rad_filter import rad_filter
from plaster.run.sim_v2.sim_v2_params import (
    DyePepType,
    DytIndexType,
    DytWeightType,
    RadType,
)
from plaster.run.sim_v2.sim_v2_result import (
    CountType,
    IndexType,
    RowKType,
    ScoreType,
    WeightType,
)
from plaster.tools.c_common import c_common_tools
from plaster.tools.c_common.c_common_tools import CException, Tab
from plaster.tools.log.log import debug
from plaster.tools.schema import check
from plaster.tools.utils import utils
from plumbum import FG, local

c_nn_v2_path = local.path("/erisyon/plaster/plaster/run/nn_v2/c")


class NNV2ContextOutputFields:
    col_names = (
        "dyt_i",
        "pep_i",
        "dyt_score",
        "score",
        "k",
        "logp_dyt",
        "logp_k",
        "dyt_21_ratio",
        "score_21_ratio",
    )

    most_likely_dyt_i = 0
    most_likely_pep_i = 1
    dyt_score = 2
    output_score = 3
    most_likely_pred_k = 4
    logp_dyt = 5
    logp_k = 6

    # The "21_ratio" stands for the ratio between the top ranking
    # score and the second-racking score
    dyt_score_21_ratio = 7
    output_score_21_ratio = 8
    n_fields = 9


class NNV2ScoringVerboseFields:
    col_names = (
        "peak_i",  #  0 test row, input radiometry
        "dyt_i",  #  1 neighbor dyetrack_i
        "weight",  #  3 target_weight, dyt_weight, how many times dyt generated by all peps
        "dist",  #  2 neighbor distance
        "s_log_z",  #  9 sum_log_z - for all cycle z_scores fitting peak rad to nrm,lognrm dist
        "p_val",  #  6 pval
        "weight_nrm",  #  4 normalized weight w.r.t all neighbors
        "penalty",  #  5 scoring penality for rare dyetracks
        "k_term",  #  7 pow(row_score_k,row_score_k_factor) term used in composite score
        "com_score",  #  8 composite_score = weight_nrm * penalty * p_val * k_term
        "dyt_score",  # 10 com score normalized relative to neighbor scores, causes high scores!
        "dyc_penalty",  # 11 penalty on winning dyetrack for high dye count
        "pv_penalty",  # 12 penalty on winning dyetrack if it did not have best p_value among candidates
        "out_score",  # 13 final output score = dyt_score * pep_score * dyc_penalty * pv_penalty
        "pep_i",  # 14 just here as diagnostic/validate we're getting the right data!
    )
    # TODO: maybe move this to c_common_tools.py in some generic form.
    @classmethod
    def emit_col_name_defines(cls, fp):
        header = "// Table columns for --scoring_verbose output\n"
        header += "\n".join(
            [f"#define SV_{name} {i}" for i, name in enumerate(cls.col_names)]
        )
        print(header, file=fp)


class NNV2Context(c_common_tools.FixupStructure):
    # fmt: off
    _fixup_fields = [
        # Input Tables
        ("train_fdyemat", Tab, RadType),
        ("train_dyepeps", Tab, DyePepType),  # 3 columns: (dyt_i, pep_i, count)
        ("radmat", Tab, RadType),
        ("radmat_filter_mask", Tab, np.bool ),
        ("ch_gain_model", Tab, np.float64 ), # 4 columns: (beta,sigma,zero_mu,zero_sigma)

        # Parameters
        ("row_k_beta", "Float64"),
        ("row_k_sigma", "Float64"),
        ("row_k_score_factor", "Float64"),

        # Options
        ("n_neighbors", "Size"),
        ("run_row_k_fit", "Bool"),
        ("run_against_all_dyetracks", "Bool"),
        ("scoring_verbose", "Bool"),
        ("scoring_verbose_cc", "Bool"),
        ("use_row_k_p_val", "Bool"),

        # Derived properties
        ("n_cols", "Size"),
        ("n_channels", "Size"),   # derived from length of GainModel.channels, see context()
        ("n_cycles", "Size"),     # n_cols / n_channels

        # Outputs

        # Main output: columns: (dyt_i, pep_i, dyt_score, score, pred_k, logp_dyt, logp_k)
        # This will always be returned even when run_against_all_dyetracks is True
        ("output", Tab, np.float64),

        # against_all_dyetracks_output is only used when run_against_all_dyetracks is True
        # in which case it is table of n_radrows with n_dyt*2 columns
        # where the columns are: (p_val, pred_k)
        ("against_all_dyetracks_output", Tab, np.float64),

        # scoring_verbose_output is only used when scoring_verbose is True
        ("scoring_verbose_output", Tab, np.float64),
        # scoring_verbose_output_cc is only used when scoring_verbose_cc is True
        ("scoring_verbose_cc_output", Tab, np.float64),

        # Internal fields
        ("_stop_requested", "Bool"),
        ("_flann_params", "struct FLANNParameters *"),
        ("_flann_index_id", "void *"),
        ("_dyt_weights", Tab, DytWeightType),
        ("_dyt_i_to_dyepep_offset", Tab, DytIndexType),
        ("_flann_index_lock", "struct pthread_mutex_t *"),
    ]
    # fmt: on

    def bad_pred_mask(self):
        return self.pred_dyt_iz == 0

    @property
    def pred_dyt_iz(self):
        return self._output[:, 0].astype(int)

    @property
    def pred_pep_iz(self):
        return self._output[:, 1].astype(int)

    @property
    def pred_dyt_scores(self):
        return self._output[:, 2]

    @property
    def pred_scores(self):
        return self._output[:, 3]

    @property
    def pred_ks(self):
        pks = self._output[:, 4]
        pks[self.bad_pred_mask()] = np.nan
        return pks

    @property
    def logp_dyt(self):
        p = self._output[:, 5]
        p[self.bad_pred_mask()] = np.nan
        return p

    @property
    def logp_k(self):
        p = self._output[:, 6]
        p[self.bad_pred_mask()] = np.nan
        return p

    @property
    def against_all_dyetrack_pvals(self):
        n_cols = self._against_all_dyetracks_output.shape[1]
        return self._against_all_dyetracks_output[:, 0 : n_cols // 3]

    @property
    def against_all_dyetrack_pred_ks(self):
        n_cols = self._against_all_dyetracks_output.shape[1]
        return self._against_all_dyetracks_output[:, n_cols // 3 : 2 * n_cols // 3]

    @property
    def against_all_dyetrack_sum_log_z_scores(self):
        n_cols = self._against_all_dyetracks_output.shape[1]
        return self._against_all_dyetracks_output[:, 2 * n_cols // 3 :]

    def to_dataframe(self):
        df = pd.DataFrame(self._output, columns=NNV2ContextOutputFields.col_names,)
        df = df.astype(
            dict(
                dyt_i=IndexType,
                pep_i=IndexType,
                dyt_score=ScoreType,
                score=ScoreType,
                k=RowKType,
                logp_dyt=ScoreType,
                logp_k=ScoreType,
                dyt_21_ratio=ScoreType,
                score_21_ratio=ScoreType,
            )
        )
        df["peak_i"] = np.arange(len(df))

        if self._radmat_filter_mask is not None:
            # Remove rows from df that did not pass RadFilter.  Note that these
            # rows contain 0's in all columns, so it's not really interesting to delay
            # this filtering.  peak_i of the kept rows addresses into the full radmat.
            assert len(self._radmat_filter_mask) == len(df)
            df = df[self._radmat_filter_mask]

        df = df.sort_values("peak_i")
        return df

    def filter_df(self, df, remove_nul_pred=False):
        mask = np.ones((len(df),), dtype=bool)
        if remove_nul_pred:
            mask &= df.dyt_i != 0
        return df[mask], mask

    def scoring_verbose_dataframe(self):
        df = pd.DataFrame(
            self._scoring_verbose_output, columns=NNV2ScoringVerboseFields.col_names,
        )
        df = df.astype(
            dict(
                peak_i=IndexType,
                dyt_i=IndexType,
                weight=CountType,
                dist=WeightType,
                s_log_z=ScoreType,
                p_val=ScoreType,
                weight_nrm=WeightType,
                penalty=WeightType,
                k_term=ScoreType,
                com_score=ScoreType,
                dyt_score=ScoreType,
                dyc_penalty=ScoreType,
                pv_penalty=ScoreType,
                out_score=ScoreType,
                pep_i=CountType,
            )
        )
        if self._radmat_filter_mask is not None:
            assert len(self._radmat_filter_mask) == len(df)
            df = df[self.radmat_filter_mask]
        df = df.sort_values(
            ["peak_i", "out_score"], ascending=[True, False]
        ).reset_index(drop=True)
        return df

    def scoring_verbose_cc_dataframe(self):
        # z_score and p_val for each column of a flattened dyetrack
        #
        # Is it necessary/advisable to have a chi_i,cycle_i in here so that e.g.
        # some merge operation doesn't mess with order of rows?
        col_names = (
            "peak_i",  # 0 test row, input radiometry
            "dyt_i",  # 1 neighbor dyetrack_i
            "z_score",  # 2 z_score
            "p_val",  # 3 p_val(z_score)
        )

        df = pd.DataFrame(self._scoring_verbose_cc_output, columns=col_names,)
        df = df.astype(
            dict(peak_i=IndexType, dyt_i=IndexType, z_score=ScoreType, p_val=ScoreType,)
        )
        if self._radmat_filter_mask is not None:
            # keep data that has passed filtering
            peak_i_keepers = np.arange(len(self._radmat_filter_mask))[
                self._radmat_filter_mask
            ]
            df = df[df.peak_i.isin(peak_i_keepers)]
        return df


def _init():
    """
    Must be called before anything else in this module
    """
    NNV2Context.struct_fixup()

    if local.env.get("PLASTER_C_COMPILE"):
        with local.cwd(c_nn_v2_path):
            fp = StringIO()
            with redirect_stdout(fp):
                print(
                    f"// This file was code-generated by nn_v2.c.nn_v2.load_lib and should be version controlled"
                )
                print()
                print("#ifndef NN_V2_H")
                print("#define NN_V2_H")
                print()
                print('#include "pthread.h"')
                print('#include "stdint.h"')
                print('#include "c_common.h"')
                print()
                NNV2ScoringVerboseFields.emit_col_name_defines(fp)
                print()
                NNV2Context.struct_emit_header(fp)
                print("#endif")

            header_file_path = "./_nn_v2.h"
            existing_h = utils.load(header_file_path, return_on_non_existing="")

            if existing_h != fp.getvalue():
                utils.save(header_file_path, fp.getvalue())

            build_dev()


# Note, this is executed at LOAD TIME!
_init()


_lib = None

MODULE_DIR = pathlib.Path(__file__).parent


def load_lib():
    global _lib
    if _lib is not None:
        return _lib

    lib = c.CDLL(MODULE_DIR / "_nn_v2.so")

    lib.sanity_check()

    lib.context_init.argtypes = [
        c.POINTER(NNV2Context),
    ]
    lib.context_init.restype = c.c_char_p

    lib.context_free.argtypes = [
        c.POINTER(NNV2Context),
    ]

    lib.classify_radrows.argtypes = [
        c.POINTER(NNV2Context),  # NNV2Context *ctx
        c_common_tools.typedef_to_ctype("Index"),  # Index radrow_start_i,
        c_common_tools.typedef_to_ctype("Size"),  # Size n_radrows,
    ]
    lib.classify_radrows.restype = c.c_char_p

    _lib = lib
    return lib


@contextmanager
def context(
    train_dyemat,
    train_dyepeps,
    radmat,
    radmat_filter_mask,
    gain_model,
    n_neighbors=8,
    run_row_k_fit=True,
    run_against_all_dyetracks=False,
    scoring_verbose=False,
    scoring_verbose_cc=False,
    use_row_k_p_val=True,
    row_k_score_factor=1.0,
):
    """
    with nn_v2.context(...) as ctx:
        zap.work_orders(do_classify_radrows, ...)
    """
    lib = load_lib()

    check.t(gain_model, GainModel)

    output_dtype = NNV2Context.tab_type("output")
    n_radrows = radmat.shape[0]
    output = np.zeros((n_radrows, NNV2ContextOutputFields.n_fields), dtype=output_dtype)

    # This is a possible place to optimize to avoid this conversion to float
    # But as it is now it is needed because the FLANN needs to lookup by float
    # so it is easier to convert is all here to RadType.
    train_fdyemat = train_dyemat.astype(RadType)
    n_dyts = train_fdyemat.shape[0]

    assert train_fdyemat.shape[1] == radmat.shape[1]
    n_cols = train_fdyemat.shape[1]
    n_channels = gain_model.n_channels
    n_cycles = n_cols // n_channels
    assert n_cycles * n_channels == n_cols

    ch_gain_model = np.array(gain_model.ch_gain_model_values())

    against_all_dyetracks_output_dtype = None
    against_all_dyetracks_output = None
    if run_against_all_dyetracks:
        against_all_dyetracks_output_dtype = NNV2Context.tab_type(
            "against_all_dyetracks_output"
        )
        against_all_dyetracks_output = np.zeros(
            (n_radrows, 3 * n_dyts), dtype=against_all_dyetracks_output_dtype
        )

    scoring_verbose_output_dtype = None
    scoring_verbose_output = None
    scoring_verbose_cc_output_dtype = None
    scoring_verbose_cc_output = None
    if scoring_verbose:
        scoring_verbose_output_dtype = NNV2Context.tab_type("scoring_verbose_output")
        scoring_verbose_output = np.zeros(
            (n_radrows * n_neighbors, len(NNV2ScoringVerboseFields.col_names)),
            dtype=scoring_verbose_output_dtype,
        )
    if scoring_verbose_cc:
        scoring_verbose_cc_output_dtype = NNV2Context.tab_type(
            "scoring_verbose_cc_output"
        )
        n_chcy = train_fdyemat.shape[1]
        scoring_verbose_cc_output = np.zeros(
            (n_radrows * n_neighbors * n_chcy, 4), dtype=scoring_verbose_cc_output_dtype
        )

    nn_v2_context = NNV2Context(
        train_fdyemat=Tab.from_mat(
            train_fdyemat, NNV2Context.tab_type("train_fdyemat")
        ),
        train_dyepeps=Tab.from_mat(
            train_dyepeps, NNV2Context.tab_type("train_dyepeps")
        ),
        radmat=Tab.from_mat(radmat, NNV2Context.tab_type("radmat")),
        radmat_filter_mask=Tab.from_mat(
            radmat_filter_mask, NNV2Context.tab_type("radmat_filter_mask")
        ),
        _radmat_filter_mask=radmat_filter_mask,
        ch_gain_model=Tab.from_mat(
            ch_gain_model, NNV2Context.tab_type("ch_gain_model")
        ),
        row_k_beta=gain_model.row_k_beta,
        row_k_sigma=gain_model.row_k_sigma,
        row_k_score_factor=row_k_score_factor,
        n_neighbors=n_neighbors,
        run_row_k_fit=run_row_k_fit,
        run_against_all_dyetracks=run_against_all_dyetracks,
        scoring_verbose=scoring_verbose,
        scoring_verbose_cc=scoring_verbose_cc,
        use_row_k_p_val=use_row_k_p_val,
        n_cols=n_cols,
        n_channels=n_channels,
        n_cycles=n_cycles,
        output=Tab.from_mat(output, output_dtype),
        _output=output,
        against_all_dyetracks_output=Tab.from_mat(
            against_all_dyetracks_output, against_all_dyetracks_output_dtype
        ),
        _against_all_dyetracks_output=against_all_dyetracks_output,
        scoring_verbose_output=Tab.from_mat(
            scoring_verbose_output, scoring_verbose_output_dtype
        ),
        _scoring_verbose_output=scoring_verbose_output,
        scoring_verbose_cc_output=Tab.from_mat(
            scoring_verbose_cc_output, scoring_verbose_cc_output_dtype
        ),
        _scoring_verbose_cc_output=scoring_verbose_cc_output,
    )
    assert (
        (-1e5 < radmat) & (radmat < 1e6)
    ).sum() > 0.5 * radmat.size, "Too many values are out of bounds for radmat"
    assert radmat.dtype == RadType

    error = lib.context_init(nn_v2_context)
    if error is not None:
        raise CException(error)

    try:
        yield nn_v2_context
    finally:
        lib.context_free(nn_v2_context)


def do_classify_radrows(nn_v2_context, radrow_start_i, n_radrows):
    lib = load_lib()
    n_actual_radrows = nn_v2_context.radmat.n_rows
    if 0 <= radrow_start_i < n_actual_radrows:
        n_radrows = min(n_actual_radrows - radrow_start_i, n_radrows)
        if radrow_start_i + n_radrows <= n_actual_radrows:
            error = lib.classify_radrows(nn_v2_context, radrow_start_i, n_radrows)
            if error is not None:
                raise CException(error)
