import numpy as np
from zest import zest
from plumbum import local
from plaster.run.sigproc_v2 import synth
from plaster.tools.test_tools.test_tools import (
    integration_before,
    integration_after,
    run_p,
)
from plaster.tools.log.log import debug


@zest.group("integration")
def zest_sigproc_v1_integration():
    # If test_folder is None it will be set by _before()
    # otherwise you can set it here to reuse a debugging folder
    test_folder = None
    # test_folder = (
    #     "/erisyon/internal/jobs_folder/_integration_tests/it_runs_sigproc_v1/1613159055"
    # )

    dim = (512, 512)
    amps_constant_val = 6000
    n_peaks = 1000
    bg_mean = 100.0
    bg_std = 30.0
    channel_scale_factor = (1.0, 0.5)
    channel_peak_width_scale_factor = (1.0, 0.8)
    psf_width = 1.5

    def _synth_field(n_cycles):
        with synth.Synth(n_channels=2, n_cycles=n_cycles, dim=dim) as s:
            peaks = (
                synth.PeaksModelGaussianCircular(n_peaks=n_peaks)
                .locs_randomize()
                .widths_uniform(psf_width)
                .channel_scale_factor(channel_scale_factor)
                .channel_peak_width_scale_factor(channel_peak_width_scale_factor)
            )
            synth.CameraModel(bg_mean=bg_mean, bg_std=bg_std)
            synth.HaloModel()
            synth.IlluminationQuadraticFalloffModel()

        return s, peaks

    def _synth_save_field(s, fl_i, source_folder):
        chcy_ims = s.render_chcy(0)

        for ch_i in range(chcy_ims.shape[0]):
            for cy_i in range(chcy_ims.shape[1]):
                np.save(
                    str(
                        source_folder
                        / f"area_{fl_i:03d}_cell_000_{ch_i:03d}nm_{cy_i:03d}.npy"
                    ),
                    chcy_ims[ch_i, cy_i],
                )

    def _synth_dyts(source_folder):
        n_fields = 3
        for fl_i in range(n_fields):
            s, peaks = _synth_field(n_cycles=4)
            peaks.dyt_random_choice(
                [[3, 2, 2, 1], [2, 1, 1, 0], [1, 0, 0, 0]], [0.5, 0.3, 0.2]
            )
            peaks.gain_constant(amps_constant_val)
            _synth_save_field(s, fl_i, source_folder)

    def _it_runs_sigproc_v1():
        job_folder = test_folder / "sigproc_v1"

        source_folder = test_folder / "sigproc_source"
        source_folder.mkdir()

        _synth_dyts(source_folder)

        run_p(
            [
                f"gen",
                f"sigproc_v1",
                f"--job={job_folder}",
                f"--sigproc_source={source_folder}",
                f"--sample=foo",
                f"--force",
            ]
        )

        run_p(["run", job_folder, "--no_progress"])

    def _it_runs_sigclass():
        """
        Test that classification can run on sigprov_v2 data
        """

        csv_file = test_folder / "protein.csv"
        job_folder = test_folder / "sigproc_v1_class"
        source_folder = test_folder / "sigproc_source"
        source_folder.mkdir()

        csv_contents = """
            Name,UniprotAC,Abundance,POI
            ACE,P12821,28,1
            ACE2,Q9BYF1,12,1
            ACTB,P60709,7,1
        """
        csv_contents = "\n".join(csv_contents.split()) + "\n"
        csv_file.write(csv_contents)

        _synth_dyts(source_folder)

        # DEBUGGING HINT: Comment out following to debug report only
        run_p(
            [
                f"gen",
                f"classify_v1",
                f"--sample=KidneyBiomarkers1",
                f"--job={job_folder}",
                f"--protein_csv={csv_file}",
                f"--label_set=C,DE",
                f"--protease=trypsin",
                f"--n_pres=1",
                f"--n_mocks=1",
                f"--n_edmans=2",
                f"--rf",
                f"--decoys=reverse",
                f"--force",
                f"--sigproc_source={source_folder}",
            ]
        )

        run_p(["run", job_folder, "--no_progress"])

    def _before():
        nonlocal test_folder
        if test_folder is None:
            test_folder = integration_before()
        test_folder = local.path(test_folder)

    def _after():
        integration_after()

    def it_runs_sigproc_v1():
        # zest randomizes order so the following ensures order...

        # WHEN DEBUGGING:
        #  * Set the test_folder to the folder that failed rather then None
        #    (indicated in the blue trace).
        #  * Comment out the following tests that you don't need to iterate on
        #    (Example, _it_runs_calib and _it_runs_sigproc_v2_with_calib has already passed
        #    and you just want to run _it_runs_sigclass over and over until it passes)

        _it_runs_sigproc_v1()
        _it_runs_sigclass()

    zest()
