import ctypes as c
import os
import pathlib
from contextlib import contextmanager, redirect_stderr, redirect_stdout
from io import StringIO

import numpy as np
from plaster.run.calib.calib import RegPSF
from plaster.run.sigproc_v2.c_radiometry.build import build_dev
from plaster.tools.c_common import c_common_tools
from plaster.tools.c_common.c_common_tools import CException, F64Arr
from plaster.tools.log.log import debug, important, prof
from plaster.tools.schema import check
from plaster.tools.utils import utils
from plaster.tools.zap import zap
from plumbum import local


class RadiometryContext(c_common_tools.FixupStructure):
    # Operates on a single field stack so that we don't have
    # to realize all fields in to memory simultaneously

    # fmt: off
    _fixup_fields = [
        ("chcy_ims", F64Arr),  # Sub-pixel aligned  (n_channels, n_cycles, height, width)
        ("locs", F64Arr),  # Sub-pixel centered  (n_peaks, 2), where 2 is: (y, x)
        ("reg_psf_samples", F64Arr),  # Reg_psf samples (n_channels, n_divs, n_divs, 3) (y, x, (sig_x, sig_y, rho))

        # Parameters
        ("n_channels", "Size"),
        ("n_cycles", "Size"),
        ("n_peaks", "Size"),
        ("n_divs", "Float64"),
        ("peak_mea", "Size"),
        ("height", "Float64"),
        ("width", "Float64"),
        ("raw_height", "Float64"),
        ("raw_width", "Float64"),

        # Outputs
        ("out_radiometry", F64Arr),  # (n_peaks, n_channels, n_cycles, 4), where 4 is: (signal, noise, snr, aspect_ratio)
    ]
    # fmt: on


c_radiometry_path = local.path("/erisyon/plaster/plaster/run/sigproc_v2/c_radiometry")


def _init():
    """
    This must be called once before any work
    """
    RadiometryContext.struct_fixup()

    if local.env.get("PLASTER_C_COMPILE"):
        with local.cwd(c_radiometry_path):
            fp = StringIO()
            with redirect_stdout(fp):
                print(
                    f"// This file was code-generated by sigproc_v2.c_radiometry.c_radiometry.load_lib and should be version controlled"
                )
                print()
                print("#ifndef RADIOMETRY_H")
                print("#define RADIOMETRY_H")
                print()
                print('#include "stdint.h"')
                print('#include "c_common.h"')
                print()
                RadiometryContext.struct_emit_header(fp)
                print("#endif")

            header_file_path = "./_radiometry.h"
            existing_h = utils.load(header_file_path, return_on_non_existing="")

            if existing_h != fp.getvalue():
                utils.save(header_file_path, fp.getvalue())

            build_dev()


# Note, this is executed at LOAD TIME!
_init()


_lib = None

MODULE_DIR = pathlib.Path(__file__).parent


class Radmat:
    # The columns of a radmat
    sig = 0
    noi = 1
    snr = 2
    asr = 3


def load_lib():
    global _lib
    if _lib is not None:
        return _lib

    lib = c.CDLL(MODULE_DIR / "_radiometry.so")
    lib.context_init.argtypes = [
        c.POINTER(RadiometryContext),
    ]
    lib.context_init.restype = c.c_char_p

    lib.context_free.argtypes = [
        c.POINTER(RadiometryContext),
    ]

    lib.radiometry_field_stack_peak_batch.argtypes = [
        c.POINTER(RadiometryContext),  # RadiometryContext context
        c_common_tools.typedef_to_ctype("Index"),  # Index peak_start_i,
        c_common_tools.typedef_to_ctype("Index"),  # Index peak_stop_i,
    ]
    lib.radiometry_field_stack_peak_batch.restype = c.c_char_p

    lib.test_interp.argtypes = [
        c.POINTER(RadiometryContext),  # RadiometryContext context
        c_common_tools.typedef_to_ctype("Float64"),  # Float64 loc_x
        c_common_tools.typedef_to_ctype("Float64"),  # Float64 loc_y
        c.c_void_p,  # Float64 *out_vals
    ]
    lib.test_interp.restype = c.c_char_p

    _lib = lib
    return lib


@contextmanager
def context(chcy_ims, locs, reg_psf: RegPSF):
    """
    with radiometry.context(...) as ctx:
        zap.work_orders(do_radiometry, ...)
    """
    lib = load_lib()

    check.array_t(chcy_ims, ndim=4, dtype=np.float64)
    n_channels, n_cycles, height, width = chcy_ims.shape

    check.array_t(locs, ndim=2, dtype=np.float64)
    check.affirm(locs.shape[1] == 2)
    n_peaks = locs.shape[0]

    check.t(reg_psf, RegPSF)
    peak_mea = reg_psf.peak_mea

    # I previously attempted to use a spline interpolator inside the
    # radiometry.c but had problems so I reverted (at least for now)
    # to a high-res sampling of the reg_psf.
    n_divs = 64
    samples = reg_psf.sample_params_grid(n_divs=n_divs)
    check.array_t(samples, ndim=4)  # (n_channels, n_divs, n_divs, 3)
    assert samples.shape[0] == n_channels

    # TODO: it would probably be a big speed improvement to pass
    # the actual rendered PSF images into the context because
    # as it is now every point inside of the radiometry c file
    # is re-rendering the PSF. This made sense when we were interpolating
    # the PSF values but if we just sample them as we do now then
    # might as well just render them once.

    out_radiometry = np.zeros((n_peaks, n_channels, n_cycles, 4), dtype=np.float64)

    reg_psf_samples = np.ascontiguousarray(samples)

    ctx = RadiometryContext(
        chcy_ims=F64Arr.from_ndarray(chcy_ims),
        locs=F64Arr.from_ndarray(locs),
        _locs=locs,
        n_channels=n_channels,
        n_cycles=n_cycles,
        n_peaks=n_peaks,
        n_divs=n_divs,
        peak_mea=peak_mea,
        height=height,
        width=width,
        raw_height=reg_psf.im_mea,
        raw_width=reg_psf.im_mea,
        n_reg_psf_samples=len(samples),
        reg_psf_samples=F64Arr.from_ndarray(reg_psf_samples),
        out_radiometry=F64Arr.from_ndarray(out_radiometry),
        _out_radiometry=out_radiometry,
    )

    error = lib.context_init(ctx)
    if error is not None:
        raise CException(error)

    try:
        yield ctx
    finally:
        lib.context_free(ctx)


def _do_radiometry_field_stack_peak_batch(
    ctx: RadiometryContext, peak_start_i: int, peak_stop_i: int
):
    """
    Worker for radiometry_field_stack() zap
    """
    lib = load_lib()

    error = lib.radiometry_field_stack_peak_batch(ctx, peak_start_i, peak_stop_i)
    if error is not None:
        raise CException(error)


def radiometry_field_stack(chcy_ims, locs, reg_psf: RegPSF):
    """
    Compute radiometry on the stack of channel and cycle images for one field

    Returns:
        output_radmat: ndarray(n_peaks, n_channels, n_cycles, (sig, noi, snr, asr))
    """
    with context(chcy_ims=chcy_ims, locs=locs, reg_psf=reg_psf,) as ctx:
        check.array_t(locs, ndim=2, dtype=np.float64)
        n_peaks = locs.shape[0]
        if n_peaks > 0:
            batches = zap.make_batch_slices(n_rows=locs.shape[0], _batch_size=100)
            with zap.Context(trap_exceptions=False, mode="thread"):
                zap.work_orders(
                    [
                        dict(
                            fn=_do_radiometry_field_stack_peak_batch,
                            ctx=ctx,
                            peak_start_i=batch[0],
                            peak_stop_i=batch[1],
                        )
                        for batch in batches
                    ]
                )

    return ctx._out_radiometry


def test_interp():
    reg_psf = RegPSF.fixture_variable()
    with context(
        chcy_ims=np.ones((1, 1, 512, 512)), locs=np.zeros((1, 2)), reg_psf=reg_psf,
    ) as ctx:
        lib = load_lib()

        diffs = []

        for y in np.linspace(0, 511, 17):
            for x in np.linspace(0, 511, 17):
                result = np.zeros((3,))
                error = lib.test_interp(ctx, x, y, result.ctypes.data_as(c.c_void_p))
                if error is not None:
                    raise CException(error)

                sig_x = reg_psf.interp_sig_x_fn[0](x, y)
                sig_y = reg_psf.interp_sig_y_fn[0](x, y)
                rho = reg_psf.interp_rho_fn[0](x, y)

                diffs += [
                    (
                        x,
                        y,
                        sig_x[0],
                        sig_y[0],
                        rho[0],
                        sig_x[0] - result[0],
                        sig_y[0] - result[1],
                        rho[0] - result[2],
                    )
                ]

        diffs = np.array(diffs)
        diffs = np.abs(diffs[:, 5:])
        return np.all(diffs < 0.03)

    return ctx._out_radiometry
