"""
Image-processing related tools.
"""

import numpy as np
from plaster.run.calib.calib import RegPSF, approximate_psf
from plaster.run.sigproc_v2.c_gauss2_fitter import gauss2_fitter
from plaster.run.sigproc_v2.c_gauss2_fitter.gauss2_fitter import (
    AugmentedGauss2Params,
    Gauss2Params,
)
from plaster.tools.image import imops
from plaster.tools.image.coord import HW, ROI, WH, XY, YX
from plaster.tools.schema import check
from plaster.tools.utils.data import one_sided_nanstd
from plaster.tools.zap import zap
from plaster.run.sigproc_v2.c_radiometry.radiometry import radiometry_field_stack
from plaster.tools.utils import data
from plaster.tools.log.log import debug, important, prof


def bandpass_filter_one_im(
    im, low_inflection, low_sharpness, high_inflection, high_sharpness
):
    """
    Use a band-pass filter to subtract out background and "bloom" which is
    the light that scatters from foreground to background.

    Note: A low_inflection of -10 effectively removes the low-pass filter
    and a high_inflection of +10 effectively removes the high-pass filter

    Values of sharpness = 50.0 are usually fine.

    Returns:
         Filtered image
    """

    # These number were hand-tuned to Abbe (512x512) and might be wrong for other
    # sizes/instruments and will need to be derived and/or calibrated.
    im = im.astype(np.float64)
    check.array_t(im, ndim=2, is_square=True, dtype=np.float64)
    low_cut = imops.generate_center_weighted_tanh(
        im.shape[0], inflection=low_inflection, sharpness=low_sharpness
    )
    high_cut = 1 - imops.generate_center_weighted_tanh(
        im.shape[0], inflection=high_inflection, sharpness=high_sharpness
    )
    filtered_im = imops.fft_filter_with_mask(im, mask=low_cut * high_cut)

    # The filters do not necessarily create a zero-centered background so
    # not remove the median to pull the background to zero.
    filtered_im -= np.median(filtered_im)

    # The bg_std is used later for tuning the peak finder.
    # Once I convert full to band-pass filter then this can just be eliminated
    # because I think it will be a constant. For now, I'm keeping
    # backward compatibility with bg_estimate_and_remove and setting
    # the constant here.
    bg_std = 3.0 * data.one_sided_nanstd(
        filtered_im.flatten(), mean=0.0, negative_side=True
    )

    return filtered_im, bg_std


def pixel_peak_find_one_im(im, approx_psf):
    """
    Peak find on a single image with 1 pixel accuracy.
    (compare to subpixel_peak_find_one_im)

    Arguments:
        im: the image to peak find
        approx_psf: An estimated PSF search kernel

    Returns:
        locs: ndarray (n_peaks_found, 2) where the 2 is in (y,x) order
    """
    from skimage.feature import peak_local_max  # Defer slow import

    std = one_sided_nanstd(im.flatten())

    # Tuning thresh:
    #  I initially had it set at 2 * std.
    #  Later, in synthetic 2 count data without a bandpass filter
    #  I found that I was losing too many 1 counts so I tried 1 * std
    #  but I found that when I put the band-pass back in that 2 * std
    #  seemed right again.  We probably should find an objective way
    #  to determine this.
    #  I think the issue is that this is measured in stdevs from the
    #  one-sided image. So in tests if you put in an image that is
    # noise free then it can't find the std which it estimates as
    # some tiny number and thus the treshold is tiny and it will
    # find WAY too many spurious peaks

    thresh = 2 * std

    cim = imops.convolve(np.nan_to_num(im, nan=float(np.nanmedian(im))), approx_psf)

    # CLEAN the edges
    # ZBS: Added because there were often edge effect from the convolution
    # that created false stray edge peaks.
    imops.edge_fill(cim, approx_psf.shape[0])

    # The background is well-described by the the histogram centered
    # around zero thanks to the fact that im and kern are expected
    # to be roughly zero-centered. Therefore we estimate the threshold
    # by using the samples less than zero cim[cim<0]
    if (cim < 0).sum() > 0:
        cim[cim < thresh] = 0
        return peak_local_max(cim, min_distance=2, threshold_abs=thresh)
    else:
        return np.zeros((0, 2))


def _subpixel_peak_find_one_im(im, peak_dim, locs):
    """
    This is a subtle calculation.

    locs is given as an *integer* position (only has pixel accuracy).
    We then extract out a sub-image using an *integer* half width.
    Peak_dim is typically odd. Suppose it is (11, 11)
    That makes half_peak_mea_i be 11 // 2 = 5

    Suppose that a peak is at (17.5, 17.5).

    Suppose that peak was found a (integer) location (17, 17)
    which is within 1 pixel of its center as expected.

    We extract the sub-image at (17 - 5, 17 - 5) = (12:23, 12:23)

    The Center-of-mass calculation should return (5.5, 5.5) because that is
    relative to the sub-image which was extracted

    We wish to return (17.5, 17.5). So that's the lower left
    (17 - 5) of the peak plus the COM found.
    """
    check.array_t(locs, dtype=int)
    assert peak_dim[0] == peak_dim[1]
    half_peak_mea_i = peak_dim[0] // 2
    lower_left_locs = locs - half_peak_mea_i
    com_per_loc = np.zeros(locs.shape)
    for loc_i, loc in enumerate(lower_left_locs):
        peak_im = imops.crop(im, off=YX(loc), dim=peak_dim, center=False)
        com_per_loc[loc_i] = imops.com(peak_im ** 2)
    return lower_left_locs + com_per_loc


def subpixel_peak_find_one_im(im, approx_psf):
    """
    First find peaks with pixel accuracy and then go back over each
    one and use the center of mass method to sub-locate them
    """
    locs = pixel_peak_find_one_im(im, approx_psf).astype(int)
    return _subpixel_peak_find_one_im(im, HW(approx_psf.shape), locs)


def _do_fg_estimate_one_field_one_channel(im, reg_psf, bandpass_kwargs, fl_i, ch_i):
    """
    Worker for fg_estimate. See fg_estimate for other details.

    For each peak that is found in the image a CIRCLE is written
    into two buffers. One buffer is the VALUE that was found of that
    peak and the other is a count.

    The values and the counts from all fields are combined and
    used to average.

    The size of the circle is arbitrary -- it is just a way to
    sort of splat the peak estimate into the general region
    """
    check.array_t(im, ndim=2)
    dim = im.shape

    # ALLOCATE a circle mask to use to splat in the values
    circle_radius = 12
    circle = imops.generate_circle_mask(circle_radius).astype(float)

    # BANDPASS filter
    filt_im, _ = bandpass_filter_one_im(im, **bandpass_kwargs)

    # FIND peaks
    locs = pixel_peak_find_one_im(filt_im, approximate_psf())

    # RADIOMETRY
    filt_im = np.ascontiguousarray(filt_im)
    radmat = radiometry_field_stack(
        filt_im[None, None, :, :], locs=locs.astype(float), reg_psf=reg_psf,
    )

    # ALLOCATE two accumulators: one for the signals and one for the counts
    fg = np.zeros(dim)
    cnt = np.zeros(dim)

    # radmat is: (n_peaks, n_channels, n_cycles, (sig, noi, snr, asr))
    # But in this case, the channel has been plucked out so it is always 0
    signals = radmat[:, 0, 0, 0]

    # FIND outliers
    if not np.all(np.isnan(signals)):
        low, high = np.nanpercentile(signals, (10, 90))

        # SPLAT circles of the intensity of the signal into an accumulator
        c = 0
        for loc, sig in zip(locs, signals):
            if low <= sig <= high:
                c += 1
                imops.accum_inplace(fg, sig * circle, loc, center=True)
                imops.accum_inplace(cnt, circle, loc, center=True)

    return fg, cnt


def fg_estimate(flch_ims, reg_psf: RegPSF, bandpass_kwargs):
    """
    Estimate the foreground illumination averaged over every field for
    each channel on cycle zero.

    This has a chicken-and-egg quality to it because we need to extract
    the foreground in order to estimate it! To resolve the chicken-and-egg
    we run the minimal peak finder on only cycle zero.

    Thus we:
        Remove the background
        Find the peaks
        Compute radiometry
        Keep the reasonable radiometry (ie percentile 10-90)
        Splat the radiometry as circles into a new image
        Make a regional summary

    Arguments:
        flch_ims: (Field, Channel, mea, mea)  (Of cycle 0)
        reg_psf: Regional PSF
        bandpass_kwargs: Args for bandpass

    Returns:
        ch_mean_ims: ndarray(n_channels, mea, mea)
            IMPORTANT: MAY CONTAIN NANS!
    """

    n_fields = flch_ims.shape[0]
    n_channels = flch_ims.shape[1]
    dim = flch_ims.shape[-2:]

    est_fg_im = np.zeros((n_channels, *dim))

    for ch_i in range(n_channels):
        fg_sum = np.zeros(dim)
        cnt_sum = np.zeros(dim)

        # ZBSL: I don't like this pattern but it will do for now.
        # I'm having to pluck out the reg psf per channel to fool
        # later code that is expecting multi-channel buffers.
        # This needs to be refactored
        ch_reg_psf = RegPSF.from_reg_psf_for_one_channel(reg_psf, ch_i)

        with zap.Context(trap_exceptions=False):
            field_fgs = zap.work_orders(
                [
                    dict(
                        fn=_do_fg_estimate_one_field_one_channel,
                        im=flch_ims[fl_i, ch_i],
                        reg_psf=ch_reg_psf,
                        bandpass_kwargs=bandpass_kwargs,
                        fl_i=fl_i,
                        ch_i=ch_i,
                    )
                    for fl_i in range(n_fields)
                ],
            )

        # field_fgs is a list at this point and I don't want
        # to force it into an array so I avoid the numpy commands here
        for fg, cnt in field_fgs:
            fg_sum += fg
            cnt_sum += cnt

        # Fill nan into all places that had no counts
        fg_sum[cnt_sum == 0] = np.nan

        # RETURN the average over the samples (fg / cnt)
        est_fg_im[ch_i] = fg_sum / cnt_sum

    return est_fg_im


def fit_peaks_one_im(im, locs, ch_i, reg_psf: RegPSF):
    """
    Run the fast Gauss2 fitter

    Arguments:
        im: A single image
        locs: The locations where to run the fitter
        ch_i: Channel to use in the reg_psf
        reg_psf: RegionalPSF (contains the PSF for all channels)

    Returns:
        fit_params: ndarray(n_locs, , AugmentedGauss2Params.N_FULL_PARAMS)
    """
    assert isinstance(reg_psf, RegPSF)

    reg_yx = np.clip(
        np.floor(reg_psf.n_divs * locs / im.shape[0]).astype(int),
        a_min=0,
        a_max=reg_psf.n_divs - 1,
    )

    n_locs = len(locs)
    guess_params = np.zeros((n_locs, AugmentedGauss2Params.N_FULL_PARAMS))

    # COPY over parameters by region for each peak
    guess_params[:, Gauss2Params.SIGMA_X] = reg_psf.get_params(
        ch_i, reg_yx[:, 0], reg_yx[:, 1], RegPSF.SIGMA_X,
    )
    guess_params[:, Gauss2Params.SIGMA_Y] = reg_psf.get_params(
        ch_i, reg_yx[:, 0], reg_yx[:, 1], RegPSF.SIGMA_Y,
    )
    guess_params[:, Gauss2Params.RHO] = reg_psf.get_params(
        ch_i, reg_yx[:, 0], reg_yx[:, 1], RegPSF.RHO,
    )

    # CENTER
    guess_params[:, Gauss2Params.CENTER_X] = reg_psf.peak_mea / 2
    guess_params[:, Gauss2Params.CENTER_Y] = reg_psf.peak_mea / 2

    # Pass zero to amp and offset to force the fitter to make its own guess
    guess_params[:, Gauss2Params.AMP] = 0.0
    guess_params[:, Gauss2Params.OFFSET] = 0.0

    fit_params, _ = gauss2_fitter.fit_image(im, locs, guess_params, reg_psf.peak_mea)
    return fit_params
