"""
This is the Signal Processor that extracts data from images from the fluoro-sequencer microscope.

Nomenclature
    Field
        One position of the X/Y stage
    Channel
        One wavelength of measured light
    Cycle
        One chemical cycle (Pre, Mock or Edman)
    Anomaly
        An area of an image that has a problem (dust, etc)
    Raw image
        Unmodified images from the scope
    Regional
        When a parameter is varies spatially
    Balance image
        A raw image scaled to compensate for regional uneven illumination
        and for differences between channels.
    Aligned field stack
        The scope stage is not perfect and does not return to exactly the same position
        each cycle, a computational alignment correction finds the optimal X/Y translation.
    ROI
        A Region Of Interest
    Intersection ROI
        In an aligned field stack, the Intersection ROI is the set
        of pixels that are in every cycle.  Ie, typically smaller than the
        dimensions of the raw images.
    Composite image
        When one or more of the channels/cycles for a field are stacked
    Fiducial images
        An image that is intended only to enhance the alignment or
        peak finding algorithm. These images are temporary and discarded after use.
    Peak/Loc/Spot
        A Peak, LOC-action, or Spot found in the image that presumably
        is generated by a single molecule.
    Radmat (aka "Radiometry Matrix")
        A matrix such that each row is a peak and each column is a measurement of brightness
        for each channel/cycle.
        Sometimes stored in (n_peaks, n_channels, n_cycles)
        Sometimes stored flatten as (n_peaks, n_channels * n_cycles)
    Radrow
        A single row (cooresponding to a single peak) of a radmat.
    cy_ims
        A set of images through all cycles for one field/channel.
    chcy_ims
        A set of images for all channel/cycles for one field.
    flchcy_ims
        A set of images for all field/channel/cycles.


Calibration-Related Components
    Sigproc Calibration is a notebook activity until we can automated it well.
    It records a Calibration object that contains:
        * regional_illumination_balance
        * regional_bg_mean
        * regional_bg_std
        * regional_psf_zstack
        * zstack_depths

V2 flow:
    0. Load calibration
        Compare the subject-id, brightness settings in the tsv files
        with what was in the calibration.
    1. Import balanced images
        Re-orders images in to output channel order
        Regionally balances images given calib
            (subtract background and scale regionally by the balance map)
        Channel equalize
            (Scale so all channels are the same strength)
    2. Mask anomalies
        Write nan into anomalies
    3. Align cycles
        Finds translations per cycles to align with cycle 0
    4. Composite with alignment offsets
        Discards pixels that are not in every cycle.
    5. Find peaks
    6. Radiometry
    7. Remove empties


TASKS:
    * Tune the size of the kernels used (1.5)
        Also, the peak_find has a similar hard-coded value

    * A general "anomaly" report section would be useful
        Frames that were very bad
        Frames where the SNR is really bad
        General histograms of quality, anomalies, and SNR

    * Examine each cycle and fit best z-depth of the PSFs
      and use that for the radiometry of that cycle

    * _compute_channel_weights needs to change over to
      calibration-time computation.

"""

import numpy as np
import pandas as pd
import plaster.run.sigproc_v2.psf
import plaster.run.sigproc_v2.reg_psf
from scipy.spatial.distance import cdist
from munch import Munch
from plaster.run.calib.calib import Calib
from plaster.run.sigproc_v2 import im_proc, psf
from plaster.run.sigproc_v2 import sigproc_v2_common as common
from plaster.run.sigproc_v2.sigproc_v2_result import SigprocV2Result
from plaster.run.sigproc_v2.c_gauss2_fitter.gauss2_fitter import Gauss2FitParams
from plaster.run.calib.calib import Calib, RegPSF, RegIllum, CalibIdentity
from plaster.tools.image import imops
from plaster.tools.schema import check
from plaster.tools.zap import zap
from plaster.tools.utils import utils
from plaster.run.sigproc_v2.c_radiometry.radiometry import radiometry_field_stack
from plaster.tools.utils.vms import vms_prof
from plaster.tools.zlog import zlog
from plaster.tools.zlog.zlog import spy
from logging import getLogger

log = getLogger(__name__)


# Calibration
# ---------------------------------------------------------------------------------------------


def _calibrate(
    ims_import_result, sigproc_v2_params, calib_identity: CalibIdentity, progress
):
    """
    Extract a PSF and extract illumination balance.

    Returns:
        calib, with new records added
    """
    check.t(calib_identity, CalibIdentity)

    calib = Calib()

    # Select fields
    # -----------------------------------

    field_iz = None
    if sigproc_v2_params.n_fields_limit is None:
        # Use quality metrics
        q = ims_import_result.qualities()
        if len(q) > 0:
            q_by_field = q.groupby("field_i").quality.mean()
            if len(q_by_field) > 20:
                # If we have lots of fields then OK to take only the best ones
                # Otherwise use all fields for calibration
                med_field_q = np.median(q_by_field.values)
                field_iz = q_by_field[q_by_field > med_field_q].index.values
    else:
        field_iz = list(range(sigproc_v2_params.n_fields_limit))

    if field_iz is None:
        field_iz = list(range(ims_import_result.n_fields))

    n_channels = ims_import_result.n_channels

    # PSF
    # -----------------------------------

    ch_reg_psfs = []
    for ch_i in range(0, n_channels):
        ch_reg_psfs += [
            psf.psf_fields_one_channel(
                ims_import_result, sigproc_v2_params, field_iz, ch_i
            )
        ]
        if progress:
            progress(ch_i + 1, n_channels)

    reg_psf = RegPSF.from_channel_reg_psfs(ch_reg_psfs)
    calib.add_reg_psf(reg_psf, calib_identity=calib_identity)

    # Regional Illumination
    # -----------------------------------

    bandpass_kwargs = dict(
        low_inflection=sigproc_v2_params.low_inflection,
        low_sharpness=sigproc_v2_params.low_sharpness,
        high_inflection=sigproc_v2_params.high_inflection,
        high_sharpness=sigproc_v2_params.high_sharpness,
    )

    im_mea = ims_import_result.dim

    flch_ims = ims_import_result.ims[field_iz, :, 0]
    ch_mean_ims = im_proc.fg_estimate(flch_ims, calib.reg_psf(), bandpass_kwargs)

    n_divs = 7  # Lowered from 10 as it seemed to just fit too much noise
    reg_illum = RegIllum(n_channels=n_channels, im_mea=im_mea, n_divs=n_divs)

    for ch_i in range(n_channels):
        reg_bal = imops.region_map(ch_mean_ims[ch_i], utils.np_safe_nanmedian, n_divs)

        check.array_t(reg_bal, shape=(n_divs, n_divs))
        if np.any(np.isnan(reg_bal)):
            # Unable to obtain reg bal for some area, use identity
            log.info(
                "Some areas of the regional balance were not detected and "
                "the entire regional balance was ignored by setting to ones."
            )
            reg_bal = np.ones_like(reg_bal)

        assert np.all(~np.isnan(reg_bal))
        reg_illum.set_ch(ch_i, reg_bal)

    calib.add_reg_illum(reg_illum, calib_identity=calib_identity)
    return calib


# Analyze Functions
# -------------------------------------------------------------------------------


def _analyze_step_1_import_balanced_images(chcy_ims, sigproc_params, calib):
    """
    Import channels and order them into the output order
    (every input channel is not necessarily used).

    Returns:
        dst_filtered_chcy_ims: Balanced and band-pass filtered
        dst_unfiltered_chcy_ims: Balanced but not band-pass filtered

    Notes:
        * Because the background is subtracted, the returned images may contain negative values.
    """
    n_channels, n_cycles = chcy_ims.shape[0:2]
    dim = chcy_ims.shape[-2:]
    dst_filt_chcy_ims = np.zeros((n_channels, n_cycles, *dim))
    dst_unfilt_chcy_ims = np.zeros((n_channels, n_cycles, *dim))

    # Removed ch_bal 18 Mar 2021 because it cases such as
    # angela/d_esn_2021_03_17_c_00_yoda012_s2_v1 it really broke things
    # and I don't think that it is yet adding much.

    # ch_bal = calib.reg_illum().channel_balance()

    # Per-frame background estimation and removal
    n_channels, n_cycles = chcy_ims.shape[0:2]
    for ch_i in range(n_channels):
        bal_im = calib.reg_illum().interp(ch_i)

        assert bal_im.sum() > 0.0, "Sanity check"

        for cy_i in range(n_cycles):
            im = np.copy(chcy_ims[ch_i, cy_i])
            if sigproc_params.run_regional_balance:
                im *= bal_im

            # im *= ch_bal[ch_i]

            filtered_im, bg_std = im_proc.bandpass_filter_one_im(
                im,
                low_inflection=sigproc_params.low_inflection,
                low_sharpness=sigproc_params.low_sharpness,
                high_inflection=sigproc_params.high_inflection,
                high_sharpness=sigproc_params.high_sharpness,
            )

            dst_unfilt_chcy_ims[ch_i, cy_i, :, :] = im
            dst_filt_chcy_ims[ch_i, cy_i, :, :] = filtered_im

    return dst_filt_chcy_ims, dst_unfilt_chcy_ims


'''
Removed temporarily because this function needs signficant tuning
and for now it is easier to remove whole fields instead of partial fields
Also, this should be moved to FG or BG as it is really a helper

def _analyze_step_2_mask_anomalies_im(im, den_threshold=300):
    """
    Operates on pre-balanced images.
    The den_threshold of 300 was found empirically on Val data

    Sets anomalies to nan
    """
    import skimage.transform  # Defer slow imports
    import cv2

    check.array_t(im, is_square=True)

    # SLICE into square using numpy-foo by reshaping the image
    # into a four-dimensional array can then by np.mean on the inner dimensions.
    sub_mea = 4  # Size of the sub-sample region
    im_mea, _ = im.shape

    squares = im.reshape(im_mea // sub_mea, sub_mea, im_mea // sub_mea, sub_mea)
    # At this point, im is now 4-dimensional like: (256, 2, 256, 2)
    # But we want the small_dims next to each other for simplicity so swap the inner axes
    squares = squares.swapaxes(1, 2)
    # Now squares is (256, 256, 2, 2.)

    # squares is like: 256, 256, 2, 2. So we need the mean of the last two axes
    squares = np.mean(squares, axis=(2, 3))

    bad_mask = (squares > den_threshold).astype(float)

    # EXPAND the bad areas by erosion and dilate.
    # Erosion gets rid of the single-pixel hits and dilation expands the bad areas
    kernel = np.ones((3, 3), np.uint8)
    mask = cv2.erode(bad_mask, kernel, iterations=1)
    mask = cv2.dilate(mask, kernel, iterations=3)

    scale = im.shape[0] // mask.shape[0]

    full_size_mask = skimage.transform.rescale(
        mask, scale=scale, multichannel=False, mode="constant", anti_aliasing=False
    ).astype(bool)

    # FIND rect contours of bad areas
    contours, hierarchy = cv2.findContours(
        full_size_mask.astype("uint8"), cv2.RETR_LIST, cv2.CHAIN_APPROX_SIMPLE
    )
    bad_rects = [cv2.boundingRect(cnt) for cnt in contours]
    im = im.copy()
    for rect in bad_rects:
        imops.fill(im, loc=XY(rect[0], rect[1]), dim=WH(rect[2], rect[3]), val=np.nan)

    return im
'''


def descent_quantities(im_A, im_B, sh_x, sh_y):
    """
    Based on John Haven Davis code.
    """
    mea = im_A.shape[0]
    assert im_A.shape[0] == im_B.shape[0]
    assert im_A.shape[1] == im_B.shape[1]
    assert im_A.shape[0] == im_A.shape[1]

    rng = np.arange(-(mea - 1) // 2, (mea + 1) // 2, 1)
    i, j = np.meshgrid(rng, rng)

    phasor = np.exp(-2.0 * complex(0.0, 1.0) * np.pi * (i * sh_x + j * sh_y) / mea)

    phasor_d_sh_x = np.multiply(phasor, (-2.0 * complex(0.0, 1.0) * np.pi * i / mea))
    phasor_d_sh_y = np.multiply(phasor, (-2.0 * complex(0.0, 1.0) * np.pi * j / mea))

    phasor_d_sh_x_d_sh_x = np.multiply(
        phasor_d_sh_x, (-2.0 * complex(0.0, 1.0) * np.pi * i / mea)
    )
    phasor_d_sh_x_d_sh_y = np.multiply(
        phasor_d_sh_x, (-2.0 * complex(0.0, 1.0) * np.pi * j / mea)
    )
    phasor_d_sh_y_d_sh_y = np.multiply(
        phasor_d_sh_y, (-2.0 * complex(0.0, 1.0) * np.pi * j / mea)
    )

    freq_dom_B = np.fft.fftshift(np.fft.fft2(im_B))
    freq_dom_shifted = np.multiply(freq_dom_B, phasor)
    freq_dom_shifted_d_sh_x = np.multiply(freq_dom_B, phasor_d_sh_x)
    freq_dom_shifted_d_sh_y = np.multiply(freq_dom_B, phasor_d_sh_y)
    freq_dom_shifted_d_sh_x_d_sh_x = np.multiply(freq_dom_B, phasor_d_sh_x_d_sh_x)
    freq_dom_shifted_d_sh_y_d_sh_y = np.multiply(freq_dom_B, phasor_d_sh_y_d_sh_y)
    freq_dom_shifted_d_sh_x_d_sh_y = np.multiply(freq_dom_B, phasor_d_sh_x_d_sh_y)

    shifted_B = np.real(np.fft.ifft2(np.fft.ifftshift(freq_dom_shifted)))
    shifted_B_d_sh_x = np.real(np.fft.ifft2(np.fft.ifftshift(freq_dom_shifted_d_sh_x)))
    shifted_B_d_sh_y = np.real(np.fft.ifft2(np.fft.ifftshift(freq_dom_shifted_d_sh_y)))
    shifted_B_d_sh_x_d_sh_x = np.real(
        np.fft.ifft2(np.fft.ifftshift(freq_dom_shifted_d_sh_x_d_sh_x))
    )
    shifted_B_d_sh_y_d_sh_y = np.real(
        np.fft.ifft2(np.fft.ifftshift(freq_dom_shifted_d_sh_y_d_sh_y))
    )
    shifted_B_d_sh_x_d_sh_y = np.real(
        np.fft.ifft2(np.fft.ifftshift(freq_dom_shifted_d_sh_x_d_sh_y))
    )

    corr = np.multiply(im_A, shifted_B)
    corr_d_sh_x = np.multiply(im_A, shifted_B_d_sh_x)
    corr_d_sh_y = np.multiply(im_A, shifted_B_d_sh_y)

    corr_d_sh_x_d_sh_x = np.multiply(im_A, shifted_B_d_sh_x_d_sh_x)
    corr_d_sh_y_d_sh_y = np.multiply(im_A, shifted_B_d_sh_y_d_sh_y)
    corr_d_sh_x_d_sh_y = np.multiply(im_A, shifted_B_d_sh_x_d_sh_y)

    obj = corr.sum()

    obj_d_sh_x = corr_d_sh_x.sum()
    obj_d_sh_y = corr_d_sh_y.sum()

    obj_d_sh_x_d_sh_x = corr_d_sh_x_d_sh_x.sum()
    obj_d_sh_y_d_sh_y = corr_d_sh_y_d_sh_y.sum()
    obj_d_sh_x_d_sh_y = corr_d_sh_x_d_sh_y.sum()

    gradient = np.array([obj_d_sh_x, obj_d_sh_y])
    hessian = np.array(
        [[obj_d_sh_x_d_sh_x, obj_d_sh_x_d_sh_y], [obj_d_sh_x_d_sh_y, obj_d_sh_y_d_sh_y]]
    )

    return obj, gradient, hessian


def _subpixel_align_one_im(fix_im, sft_im):
    sft_x = 0.0
    sft_y = 0.0
    for step in range(10):
        objective, gradient, hessian = descent_quantities(fix_im, sft_im, sft_x, sft_y)
        delta = np.linalg.solve(hessian, gradient)
        # TODO: Optimize, it might be worth scaling delta and setting the
        #       exit criteria to some value other than 1e-6
        sft_x += -delta[0]
        sft_y += -delta[1]
        if delta[0] ** 2 < 1e-6 ** 2 and delta[1] ** 2 < 1e-6 ** 2:
            break
    else:
        log.info("Failed to converge on sub pixel alignment.")
        return (np.nan, np.nan)

    # JHD uses opposite shifting notation so the following is negated.
    return -sft_y, -sft_x


def _sub_pixel_align_cy_ims(cy_ims, bounds=None):
    # Get within on pixel
    pixel_offsets, pixel_aligned_cy_ims = imops.align(
        cy_ims, return_shifted_ims=True, bounds=bounds
    )

    # Then use gradient descent to get sub-pixel accuracy
    sub_pixel_offsets = np.zeros((pixel_aligned_cy_ims.shape[0], 2))
    pixel_aligned_cy0_im = pixel_aligned_cy_ims[0]
    for cy_i, pixel_aligned_cy_im in enumerate(pixel_aligned_cy_ims):
        if cy_i == 0:
            continue
        _dy, _dx = _subpixel_align_one_im(pixel_aligned_cy0_im, pixel_aligned_cy_im)
        sub_pixel_offsets[cy_i, :] = (_dy, _dx)

    return pixel_offsets + sub_pixel_offsets


def _align_image_stack(ims, bounds=None):
    """
    Align a stack of ims by generating simplified fiducials for each image.

    If bounds, do not allow alignment > bounds

    Returns:
        aln_offsets: ndarray(n_cycles, 2); where 2 is (y, x)
    """

    approx_psf = plaster.run.calib.calib.approximate_psf()

    fiducial_ims = np.zeros_like(ims)
    for i, im in enumerate(ims):
        im = im.astype(np.float64)
        if not np.all(np.isnan(im)):
            med = float(np.nanmedian(im))
        else:
            med = 0
        im = np.nan_to_num(im, nan=med)
        fid_im = imops.convolve(im, approx_psf - np.mean(approx_psf))
        fid_im -= np.median(fid_im)
        imops.edge_fill(fid_im, 20)
        fiducial_ims[i, :, :] = fid_im

    return _sub_pixel_align_cy_ims(fiducial_ims, bounds=bounds)


def _analyze_step_3_align_chcy_ims(chcy_ims, channel_align_bounds):
    """
    In some cases the stage can move slightly between channels on the Nikon scope
    presumably because there's a mechanical turret that rotates to select filters.
    This will likely not be an issue on final scope as all channels will
    be acquired simultaneously.

    Because of this movement, we align each channel stack independently
    and then align all the channel stacks.
    But, we limit the channel-to-channel alignment so that it can't
    go crazy.
    """
    check.array_t(chcy_ims, ndim=4)
    n_channels, n_cycles, h, w = chcy_ims.shape

    # ALIGN cycle 0 of each channel relative to channel 0
    # Prevent this from going too crazy by bounding
    chcy0_ims = chcy_ims[:, 0]
    ch_offsets = _align_image_stack(chcy0_ims, bounds=channel_align_bounds)

    # ALIGN each channel stack independently
    offsets_each_channel = np.zeros((n_channels, n_cycles, 2))
    for ch_i, chicy_ims in enumerate(chcy_ims):
        offsets_each_channel[ch_i, :, :] = _align_image_stack(chicy_ims)

    # COMBINE the alignments from each channel stack using the
    # cycle 0 channel-to-channel alignments
    chcy_offsets = offsets_each_channel + ch_offsets[:, None, :]
    return chcy_offsets, offsets_each_channel, ch_offsets


def _analyze_step_4_align_stack_of_chcy_ims(chcy_ims, chcy_offsets):
    """
    Given the alignment offsets for each channel and cycle,
    create a new image stack that has the dimensions of the
    intersection ROI (ie the overlapping
    region that contains pixels from all cycles)

    Returns:
        A newly allocated ndarray(n_channels, n_cycles, ROI)
        where ROI is the region of interest determined by the
        pixels that all cycles have in common

    Notes:
        The returned image is likely smaller than the chcy_ims shape.
    """
    check.array_t(chcy_ims, ndim=4)
    n_channels, n_cycles, h, w = chcy_ims.shape

    # STACK all channel offsets to compute the ROI for the whole stack
    all_channel_aln_offsets = np.vstack(chcy_offsets)
    raw_dim = chcy_ims.shape[-2:]
    roi = imops.intersection_roi_from_aln_offsets(all_channel_aln_offsets, raw_dim)
    roi_dim = (roi[0].stop - roi[0].start, roi[1].stop - roi[1].start)
    roi_mea = min(*roi_dim)

    # RESAMPLE the original images with sub-pixel accuracy into a final buffer
    aligned_chcy_ims = np.zeros((n_channels, n_cycles, roi_mea, roi_mea))

    for ch_i in range(n_channels):
        for cy_i, offset in zip(range(n_cycles), chcy_offsets[ch_i]):
            # Sub-pixel shift the square raw images using phase shifting
            # (This must be done with the original square images)
            im = chcy_ims[ch_i, cy_i]
            shifted_im = imops.fft_sub_pixel_shift(im, -offset)

            # Now that it is shifted we pluck out the ROI into the destination
            aligned_chcy_ims[ch_i, cy_i, 0:roi_mea, 0:roi_mea] = shifted_im[
                slice(roi[0].start, roi[0].start + roi_mea),
                slice(roi[1].start, roi[1].start + roi_mea),
            ]

    return aligned_chcy_ims


def _analyze_step_5_find_peaks(chcy_ims, kernel, cycle_i):
    """
    Step 5: Peak find on combined channels

    Previous verion of this code depended on the channels being
    balanced relative to one another. But this early-stage channel
    balancing turned out to be problematic.

    This new code instead peak finds on each channel independently
    and then reconciles the peak locations by unioning the
    lists of peaks and de-duping them using an approximate measure
    of distance as the de-dupe key.

    Returns:
        locs: ndarray (n_peaks, 2)  where the second dim is in (y, x) order
    """

    # Use more than one cycle to improve the quality of the sub-pixel estimate
    # But then discard peaks that are off after cycle 1??

    n_channels = chcy_ims.shape[0]
    locs_per_channel = []
    for ch_i in range(n_channels):
        im = chcy_ims[ch_i, cycle_i, :, :]
        try:
            locs = im_proc.subpixel_peak_find_one_im(im, kernel)
        except Exception:
            log.exception("Failure during peak find, no peaks recorded for this frame.")
            locs = np.zeros((0, 2))
        locs_per_channel += [locs]

    union_locs = np.vstack(locs_per_channel)

    # UNION and de-dupe where < 1.0(?) pixel is considered a duplicate
    dists = cdist(union_locs, union_locs, "euclidean")
    n_locs = union_locs.shape[0]
    if n_locs == 0:
        return np.zeros((0, 2))

    # Set self-distances to large so that they will not be found as "closest"
    dists[np.arange(n_locs), np.arange(n_locs)] = 1e6
    closest_i = np.argmin(dists, axis=1)
    closest_d = dists[np.arange(n_locs), closest_i]

    # Any peaks with another peak within 1 pixel is a candidate for de-dupeing
    dedupe_mask = closest_d <= 1.0

    a_iz = closest_i[dedupe_mask]
    b_iz = np.arange(n_locs)[dedupe_mask]
    c_iz = np.arange(n_locs)[~dedupe_mask]  # Those with no mate (in one channel only)

    # Of those pairs we have to keep one or the other so we take the
    # one with the lower index value.
    keep_iz = np.where(a_iz < b_iz, a_iz, b_iz)
    keep_iz = np.concatenate((keep_iz, c_iz))
    return union_locs[np.unique(keep_iz)]


def _analyze_step_6a_fitter(
    chcy_ims, locs, reg_psf: plaster.run.calib.calib.RegPSF, mask
):
    """
    Fit Gaussian.

    Arguments:
        chcy_ims: (n_channels, n_cycles, width, height)
        locs: (n_peaks, 2). The second dimension is in (y, x) order
        calib: Calibration (needed for psf)
        psf_params: The Gaussian (rho form) params for the entire PSF stack
        mask: Used to subsample the locs (where true)

    Returns:
        fitmat: ndarray(n_locs, n_channels, n_cycles, 3 + 8)
            Where the last dim is (sig, noi, asr) + (params of gaussian in rho form)

    """
    check.array_t(chcy_ims, ndim=4)
    check.array_t(locs, ndim=2, shape=(None, 2))
    check.array_t(mask, shape=(locs.shape[0],), dtype=bool)

    n_locs = len(locs)
    n_channels, n_cycles = chcy_ims.shape[0:2]

    fitmat = np.full(
        (n_locs, n_channels, n_cycles, Gauss2FitParams.N_FULL_PARAMS), np.nan
    )

    locs = locs.copy()
    locs[~mask, :] = np.nan

    for ch_i in range(n_channels):
        for cy_i in range(n_cycles):
            im = chcy_ims[ch_i, cy_i]
            fitmat[:, ch_i, cy_i, :] = im_proc.fit_peaks_one_im(im, locs, ch_i, reg_psf)

    return fitmat


def _analyze_step_6b_radiometry(chcy_ims, locs, calib):
    """
    Extract radiometry (signal and noise) from the field chcy stack.

    Arguments:
        chcy_ims: (n_channels, n_cycles, width, height)
        locs: (n_peaks, 2). The second dimension is in (y, x) order
        calib: Calibration (needed for psf)

    Returns:
        radmat: ndarray(n_locs, n_channels, n_cycles, 3)
            Where the last dim is (signal, noise, aspect_ratio)
    """
    check.array_t(chcy_ims, ndim=4)
    check.array_t(locs, ndim=2, shape=(None, 2))

    n_channels, n_cycles = chcy_ims.shape[0:2]
    reg_psf = calib.reg_psf()

    radmat = radiometry_field_stack(chcy_ims, locs=locs, reg_psf=reg_psf)

    return radmat


def _sigproc_analyze_field(
    raw_chcy_ims,
    sigproc_v2_params,
    calib,
    reg_psf: plaster.run.calib.calib.RegPSF = None,
):
    """
    Analyze one field --
        * Regional and channel balance
        * remove anomalies (temporarily removed)
        * Align cycles
        * Composite aligned
        * Peak find
        * Radiometry
        * Filtering (temporarily removed)

    Arguments:
        filt_chcy_ims: from ims_import_result
        sigproc_v2_params: The SigprocParams
        calib: calibration
    """

    # Step 1: Load the images in output channel order, balance, equalize
    # Timings:
    #   Val8_2t: 20 seconds per field, using a single core, probably IO bound
    filt_chcy_ims, unfilt_chcy_ims = _analyze_step_1_import_balanced_images(
        raw_chcy_ims.astype(np.float64), sigproc_v2_params, calib
    )

    n_cycles = filt_chcy_ims.shape[1]

    """
    Removed temporarily see _analyze_step_2_mask_anomalies_im for explanation
    # Step 2: Remove anomalies (at least for alignment)
    if sigproc_v2_params.run_anomaly_detection:
        for ch_i, cy_ims in enumerate(chcy_ims):
            chcy_ims[ch_i] = imops.stack_map(cy_ims, _analyze_step_2_mask_anomalies_im)
    """

    # Step 3: Find alignment offsets by using the mean of all channels
    # Note that this requires that the channel balancing has equalized the channel weights
    # Timings:
    #   Val8_2t: 53 seconds per field, single core for about 20 seconds and then
    #            several bursts of all cores. Presumably that early delay is load time
    #   Val10_1t debug_mode on remote_dev:
    # if sigproc_v2_params.run_aligner:
    #     aln_offsets = _align_image_stack(np.mean(filt_chcy_ims, axis=0))
    # else:
    #     aln_offsets = np.zeros((n_cycles, 2))

    # Step 4: Composite with alignment
    # Timings:
    #   Val8_2t: Each of the following two taking 14 sec (28 sec combined)
    #            Completely single core. This could probably be parallelized somehow

    chcy_offsets, offsets_each_channel, ch_offsets = _analyze_step_3_align_chcy_ims(
        filt_chcy_ims, channel_align_bounds=sigproc_v2_params.channel_align_bounds
    )

    aln_filt_chcy_ims = _analyze_step_4_align_stack_of_chcy_ims(
        filt_chcy_ims, chcy_offsets
    )
    aln_unfilt_chcy_ims = _analyze_step_4_align_stack_of_chcy_ims(
        unfilt_chcy_ims, chcy_offsets
    )

    # aln_*filt_chcy_ims is now only the shape of only intersection region so is likely
    # to be smaller than the original and not necessarily a power of 2.

    # Step 5: Peak find on combined channels
    # The goal of previous channel equalization and regional balancing is that
    # all pixels are now on an equal footing so we can now use
    # a single values for fg_thresh and bg_thresh.

    approx_psf = plaster.run.calib.calib.approximate_psf()
    if sigproc_v2_params.locs is not None:
        locs = np.array(sigproc_v2_params.locs)
        assert locs.ndim == 1
        locs = np.reshape(locs, (locs.shape[0] // 2, 2))
    else:
        locs = _analyze_step_5_find_peaks(aln_filt_chcy_ims, approx_psf, cycle_i=0)

    n_locs = len(locs)

    cy_locs = None
    if sigproc_v2_params.run_per_cycle_peakfinder:
        cy_locs = []
        for cy_i in range(n_cycles):
            cy_locs += [
                _analyze_step_5_find_peaks(aln_filt_chcy_ims, approx_psf, cycle_i=cy_i)
            ]

    # Step 6: Radiometry over each channel, cycle

    # Sample all or a sub-set of peaks for focus purposes (and also for debugging)
    if sigproc_v2_params.run_analysis_gauss2_fitter:
        mask = np.ones((n_locs,), dtype=bool)

    else:
        # Subsample peaks for fitting
        mask = np.zeros((n_locs,), dtype=bool)
        count = 250
        if n_locs > 0:
            try:
                iz = np.random.choice(
                    n_locs, count
                )  # Allow replace in case count > n_locs
                mask[iz] = 1
            except ValueError:
                pass

    fitmat = _analyze_step_6a_fitter(aln_filt_chcy_ims, locs, reg_psf, mask)

    # At moment it appears that focus adjustment does nothing under the
    # the filters and alignment -- because there is no correlation anymore
    # between peak width and brightness, so for now I'm turning off the
    # focus adjustment but leacing in the fitmat sampling for reporting purposes

    # This is taking about 2.5 seconds (0.5 of which is the interpolation of the psf)
    # Seems surprisingly slow. Something is up.
    radmat = _analyze_step_6b_radiometry(aln_filt_chcy_ims, locs, calib)

    neighborhood_stats = None
    if sigproc_v2_params.run_neighbor_stats:
        from plaster.tools.image.coord import WH, XY

        sub_mea = 31
        peak_mea = 11
        bot = sub_mea // 2 - peak_mea // 2
        top = bot + peak_mea
        lft = sub_mea // 2 - peak_mea // 2
        rgt = bot + peak_mea
        ch_i = 0
        neighborhood_stats = np.zeros((n_locs, n_cycles, 4))
        with utils.np_no_warn():
            for loc_i, (y, x) in enumerate(locs):
                x = int(x + 0.5)
                y = int(y + 0.5)
                for cy_i in range(n_cycles):
                    sub_im = imops.crop(
                        aln_filt_chcy_ims[ch_i, cy_i],
                        off=XY(x, y),
                        dim=WH(32, 32),
                        center=True,
                    ).copy()
                    sub_im[bot:top, lft:rgt] = np.nan
                    neighborhood_stats[loc_i, cy_i, 0] = np.nanmean(sub_im)
                    neighborhood_stats[loc_i, cy_i, 1] = np.nanstd(sub_im)
                    neighborhood_stats[loc_i, cy_i, 2] = np.nanmedian(sub_im)
                    a = np.nanpercentile(sub_im, [75, 25])
                    try:
                        neighborhood_stats[loc_i, cy_i, 3] = a[0] - a[1]
                    except IndexError as e:
                        neighborhood_stats[loc_i, cy_i, 3] = np.nan

    return (
        aln_filt_chcy_ims,
        aln_unfilt_chcy_ims,
        locs,
        radmat,
        chcy_offsets,
        fitmat,
        neighborhood_stats,
        cy_locs,
        offsets_each_channel,
        ch_offsets,
    )


def _do_sigproc_analyze_and_save_field(
    field_i, ims_import_result, sigproc_v2_params, sigproc_v2_result, calib
):
    """
    Analyze AND SAVE one field by calling the sigproc_v2_result.save_field()
    """

    with zlog.add_log_fields(field_i=field_i):

        chcy_ims = ims_import_result.ims[field_i]
        n_channels, n_cycles, roi_h, roi_w = chcy_ims.shape

        reg_psf = calib.reg_psf()

        (
            aln_filt_chcy_ims,
            aln_unfilt_chcy_ims,
            locs,
            radmat,
            aln_offsets,
            fitmat,
            neighborhood_stats,
            cy_locs,
            offsets_each_channel,
            ch_offsets,
        ) = _sigproc_analyze_field(chcy_ims, sigproc_v2_params, calib, reg_psf)

        mea = np.array([chcy_ims.shape[-1:]])
        bad_align_mask = aln_offsets ** 2 > (mea * 0.2) ** 2
        # if np.any(bad_align_mask):
        #     log.info(
        #         f"field {field_i} has bad alignment @(ch,cy,dim[h,w])\n"
        #         f"{np.argwhere(bad_align_mask)}\n"
        #         f"with values:\n"
        #         f"{aln_offsets[bad_align_mask]}"
        #     )

        # Assign 0 to "peak_i" in the following DF because that is the GLOBAL peak_i
        # which is not computable until all fields are processed. It will be fixed up later
        # by the SigprocV2Result helper methods
        peak_df = pd.DataFrame(
            [(0, field_i, peak_i, loc[0], loc[1]) for peak_i, loc in enumerate(locs)],
            columns=list(SigprocV2Result.peak_df_schema.keys()),
        )

        field_df = pd.DataFrame(
            [
                (
                    field_i,
                    ch_i,
                    cy_i,
                    aln_offsets[ch_i, cy_i, 0],
                    aln_offsets[ch_i, cy_i, 1],
                )
                for ch_i in range(n_channels)
                for cy_i in range(n_cycles)
            ],
            columns=list(SigprocV2Result.field_df_schema.keys()),
        )

        assert len(radmat) == len(peak_df)

        sigproc_v2_result.save_field(
            field_i,
            peak_df=peak_df,
            field_df=field_df,
            radmat=radmat,
            fitmat=fitmat,
            neighborhood_stats=neighborhood_stats,
            cy_locs=cy_locs,
            _aln_filt_chcy_ims=aln_filt_chcy_ims,
            _aln_unfilt_chcy_ims=aln_unfilt_chcy_ims,
            offsets_each_channel=offsets_each_channel,
            ch_offsets=ch_offsets,
            aln_offsets=aln_offsets,
        )


# Entrypoints
# -------------------------------------------------------------------------------


def sigproc_instrument_calib(sigproc_v2_params, ims_import_result, progress=None):
    """
    Entrypoint for Illumination and PSF calibration.
    """

    focus_per_field_per_channel = None
    calib = None

    assert sigproc_v2_params.mode == common.SIGPROC_V2_ILLUM_CALIB
    if sigproc_v2_params.mode == common.SIGPROC_V2_ILLUM_CALIB:
        assert sigproc_v2_params.instrument_identity is not None
        calib = _calibrate(
            ims_import_result,
            sigproc_v2_params,
            CalibIdentity(sigproc_v2_params.instrument_identity),
            progress,
        )

    sigproc_v2_result = SigprocV2Result(
        params=sigproc_v2_params,
        n_channels=None,
        n_cycles=None,
        n_fields=None,
        channel_weights=None,
        calib=calib,
        focus_per_field_per_channel=focus_per_field_per_channel,
    )
    sigproc_v2_result.save()
    return sigproc_v2_result


def sigproc_analyze(
    sigproc_v2_params, ims_import_result, progress, calib=None, pipeline=None
):
    """
    Entrypoint for analysis of (ie generate radiometry).
    Requires a calibration_file previously generated by sigproc_instrument_calib()
    that is referred to in sigproc_v2_params.calibration_file

    If calib is not None it over-rides the loading of a calibration_file
    (used for testing)
    """

    phase = 0
    n_phases = 1

    if False:
        # Set to True to debug quickly and reload the self.calib
        log.warning("USING DEBUG MODE CACHED CALIBRATION")
        calib = Calib.load_file("self.calib", "_self")
        assert (
            calib.has_records()
        ), "load of calib had no records. Maybe it was'nt a self calib?"

    elif sigproc_v2_params.self_calib:
        n_phases = 2
        if pipeline:
            pipeline.set_phase(phase, n_phases)
            phase += 1

        self_calib_identity = CalibIdentity("_self")
        calib = _calibrate(
            ims_import_result, sigproc_v2_params, self_calib_identity, progress
        )
        calib.save_file("self.calib")

    elif sigproc_v2_params.no_calib:
        assert sigproc_v2_params.instrument_identity is None
        assert (
            sigproc_v2_params.no_calib_psf_sigma is not None
        ), "In no_calib mode you must specify an estimated no_calib_psf_sigma"
        calib_identity = CalibIdentity("_identity")

        calib = Calib()
        arr = np.zeros((ims_import_result.n_channels, 5, 5, RegPSF.N_PARAMS))
        arr[:, :, :, RegPSF.SIGMA_X] = sigproc_v2_params.no_calib_psf_sigma
        arr[:, :, :, RegPSF.SIGMA_Y] = sigproc_v2_params.no_calib_psf_sigma
        arr[:, :, :, RegPSF.RHO] = 0.0
        reg_psf = RegPSF.from_array(im_mea=ims_import_result.dim, peak_mea=11, arr=arr)
        calib.add_reg_psf(reg_psf, calib_identity)

        reg_illum = RegIllum.identity(
            ims_import_result.n_channels, ims_import_result.dim
        )
        calib.add_reg_illum(reg_illum, calib_identity)

    elif calib is None:
        assert sigproc_v2_params.instrument_identity is not None
        calib = Calib.load_file(
            sigproc_v2_params.calibration_file, sigproc_v2_params.instrument_identity
        )
        assert calib.has_records()

    n_fields = ims_import_result.n_fields
    n_fields_limit = sigproc_v2_params.n_fields_limit
    if n_fields_limit is not None and n_fields_limit < n_fields:
        n_fields = n_fields_limit

    sigproc_v2_result = SigprocV2Result(
        params=sigproc_v2_params,
        n_fields=n_fields,
        n_channels=ims_import_result.n_channels,
        n_cycles=ims_import_result.n_cycles,
        calib=calib,
        focus_per_field_per_channel=None,
    )

    if pipeline:
        pipeline.set_phase(phase, n_phases)
        phase += 1

    approx_field_memory = (
        ims_import_result.dim ** 2  # n_pixels
        * ims_import_result.n_channels  # n_channels
        * ims_import_result.n_cycles  # n_cycles
        * 8  # sizeof(float64)
        * 2  # filtered and unfiltered
        * 5  # working space
    )

    field_list = list(range(n_fields))
    if False:
        # Use this to debug a specific field
        field_list = (20,)
        log.warning(f"USING DEBUG MODE LIMIT OF FIELD_LIST: {field_list}")

    with zap.Context(
        trap_exceptions=False, progress=progress, mem_per_workorder=approx_field_memory
    ):
        zap.work_orders(
            [
                Munch(
                    fn=_do_sigproc_analyze_and_save_field,
                    field_i=field_i,
                    ims_import_result=ims_import_result,
                    sigproc_v2_params=sigproc_v2_params,
                    sigproc_v2_result=sigproc_v2_result,
                    calib=calib,
                )
                for field_i in field_list
            ]
        )

    sigproc_v2_result.save(
        save_full_signal_radmat_npy=sigproc_v2_params.save_full_signal_radmat_npy
    )

    return sigproc_v2_result
