import numpy as np
import plaster.run.calib.calib
import plaster.run.sigproc_v2.psf
from plaster.run.sigproc_v2 import synth, im_proc
from plaster.run.sigproc_v2.c_radiometry.radiometry import radiometry_field_stack
from plaster.tools.image.coord import HW
from plaster.tools.image import imops
from plaster.tools.utils import utils
from plaster.run.calib.calib import RegPSF
from plaster.run.sigproc_v2.c_radiometry import radiometry
from zest import zest
from plaster.tools.log.log import debug


def zest_bandpass_filter_one_im():
    def it_removes_bg_by_low_cut_one_cycle():
        reg_psf = RegPSF.fixture()
        with synth.Synth(dim=(512, 512), n_cycles=1) as s:
            peaks = (
                synth.PeaksModelPSF(reg_psf, n_peaks=100)
                .amps_constant(5000.0)
                .locs_grid(pad=50)
            )
            synth.CameraModel(bg_mean=100, bg_std=10)
            synth.HaloModel()
            chcy_ims = s.render_chcy()

        im, std = im_proc.bandpass_filter_one_im(
            chcy_ims[0, 0],
            low_inflection=0.03,
            low_sharpness=50.0,
            high_inflection=0.5,
            high_sharpness=50.0,
        )

        # CHECK that it removed the bg
        assert -2.0 < np.median(im) < 2.0

        # CHECK that it did not affect the fg by much
        bg_removed_chcy_ims = im[None, None, :, :]
        bg_removed_chcy_ims = np.ascontiguousarray(
            bg_removed_chcy_ims, dtype=np.float64
        )

        radmat = radiometry_field_stack(
            bg_removed_chcy_ims, locs=peaks.locs, reg_psf=reg_psf,
        )

        assert 4900.0 < np.mean(radmat[:, 0, 0, 0]) < 5000.0
        assert 0.0 < np.std(radmat[:, 0, 0, 0]) < 100.0

    def it_removes_bg_by_low_cut():
        with synth.Synth(dim=(512, 512), n_cycles=3) as s:
            peaks = (
                synth.PeaksModelGaussianCircular(n_peaks=100)
                .widths_uniform(1.5)
                .locs_grid(pad=50)
            )
            synth.CameraModel(bg_mean=100, bg_std=10)
            synth.HaloModel()
            peaks.dyt_random_choice([[1, 1, 1], [1, 0, 1]], [0.5, 0.5]).gain_constant(
                5000
            )
            chcy_ims = s.render_chcy()

        im, std = im_proc.bandpass_filter_one_im(
            chcy_ims[0, 0],
            low_inflection=0.03,
            low_sharpness=50.0,
            high_inflection=0.5,
            high_sharpness=50.0,
        )
        median, high = np.percentile(im, (50, 99.9))
        assert np.abs(median) < 1.0
        assert np.abs(high) < 400.0

    zest()


def zest_peak_find():
    def it_finds_one_peak_sub_pixel_exactly_under_ideal_conditions():
        true_locs = np.array([[17.5, 17.5]])
        peak_im = imops.gauss2_rho_form(
            amp=1000.0,
            std_x=1.8,
            std_y=1.8,
            pos_x=true_locs[0, 1],
            pos_y=true_locs[0, 0],
            rho=0.0,
            const=0.0,
            mea=35,
        )

        pred_locs = im_proc._subpixel_peak_find_one_im(
            peak_im, HW(11, 11), true_locs.astype(int)
        )
        dists = np.linalg.norm(true_locs - pred_locs, axis=1)
        assert np.all(dists < 0.001)

    def it_finds_one_peak_sub_pixel_exactly_under_ideal_conditions_many_offsets():
        for trials in range(50):
            true_locs = np.random.uniform(-5, 5, size=(1, 2))
            true_locs += 35 / 2
            peak_im = imops.gauss2_rho_form(
                amp=1000.0,
                std_x=1.8,
                std_y=1.8,
                pos_x=true_locs[0, 1],
                pos_y=true_locs[0, 0],
                rho=0.0,
                const=0.0,
                mea=35,
            )

            pred_locs = im_proc._subpixel_peak_find_one_im(
                peak_im, HW(11, 11), true_locs.astype(int)
            )
            dists = np.linalg.norm(true_locs - pred_locs, axis=1)
            assert np.all(dists < 0.001)

    def it_find_pixel_accurate():
        bg_std = 10
        with synth.Synth(dim=(512, 512), n_cycles=3) as s:
            true_n_peaks = 100
            peaks = (
                synth.PeaksModelGaussianCircular(n_peaks=true_n_peaks)
                .amps_constant(1000)
                .widths_uniform(1.5)
                .locs_randomize()
            )
            s.zero_aln_offsets()
            synth.CameraModel(0, bg_std)
            chcy_ims = s.render_chcy()

        im, bg_std = im_proc.bandpass_filter_one_im(
            chcy_ims[0, 0],
            low_inflection=0.03,
            low_sharpness=50.0,
            high_inflection=0.5,
            high_sharpness=50.0,
        )
        kernel = plaster.run.calib.calib.approximate_psf()
        locs = im_proc.pixel_peak_find_one_im(chcy_ims[0, 0], kernel)
        n_peaks, n_dims = locs.shape
        assert n_dims == 2
        assert n_peaks > 0.80 * true_n_peaks

    def it_finds_sub_pixel_exactly_under_ideal_conditions():
        """
        Test the helper _sub_pixel_peak_find instead of sub_pixel_peak_find
        because we don't want to have to reconcile the peak ordering
        from the synth with the arbitrary order they are found by the
        peak finder
        """

        with synth.Synth(dim=(512, 512), n_cycles=3) as s:
            true_n_peaks = 100
            peaks = (
                synth.PeaksModelGaussianCircular(n_peaks=true_n_peaks)
                .amps_constant(1000)
                .widths_uniform(1.5)
                .locs_grid()
                .locs_add_random_subpixel()
            )
            s.zero_aln_offsets()
            chcy_ims = s.render_chcy()

        chcy_mean_im = np.mean(chcy_ims, axis=(0, 1))
        locs = im_proc._subpixel_peak_find_one_im(
            chcy_mean_im, HW(peaks.mea, peaks.mea), peaks.locs.astype(int)
        )
        dists = np.linalg.norm(locs - peaks.locs, axis=1)
        assert np.all(dists < 0.01)

    def it_finds_sub_pixel_well_under_typical_conditions():
        bg_std = 10
        with synth.Synth(dim=(512, 512), n_cycles=3) as s:
            true_n_peaks = 100
            peaks = (
                synth.PeaksModelGaussianCircular(n_peaks=true_n_peaks)
                .amps_constant(1000)
                .widths_uniform(1.5)
                .locs_randomize_away_from_edges()
            )
            synth.CameraModel(0, bg_std)
            s.zero_aln_offsets()
            chcy_ims = s.render_chcy()

        chcy_mean_im = np.mean(chcy_ims, axis=(0, 1))
        locs = im_proc._subpixel_peak_find_one_im(
            chcy_mean_im, HW(peaks.mea, peaks.mea), peaks.locs.astype(int)
        )
        dists = np.linalg.norm(locs - peaks.locs, axis=1)

        assert (dists < 0.1).sum() > 30
        assert (dists < 0.2).sum() > 70

    zest()


def zest_fg_estimate():
    bandpass_kwargs = dict(
        low_inflection=0.03,
        low_sharpness=50.0,
        high_inflection=0.5,
        high_sharpness=50.0,
    )

    def _gen_field_one_channel(n_peaks):
        with synth.Synth(n_channels=1, dim=(512, 512), n_cycles=1) as s:
            (
                synth.PeaksModelGaussianCircular(n_peaks=n_peaks)
                .amps_constant(2000.0)
                .widths_uniform(1.8)
                .locs_randomize()
            )
            synth.IlluminationQuadraticFalloffModel(width=0.9)
            synth.CameraModel()
            return s.render_chcy()[0, 0]

    def _gen_field_two_channel():
        with synth.Synth(n_channels=2, dim=(512, 512), n_cycles=1) as s:
            (
                synth.PeaksModelGaussianCircular(n_peaks=800)
                .amps_constant(2000.0)
                .widths_uniform(1.8)
                .locs_randomize()
                .channel_scale_factor((1.0, 0.5))
            )
            synth.IlluminationQuadraticFalloffModel(width=np.array([0.9, 5.0]))
            synth.CameraModel()
            return s.render_chcy()[:, 0]

    def it_handles_single_channel():
        ch0_ims = np.stack(
            (
                _gen_field_one_channel(800),
                _gen_field_one_channel(800),
                _gen_field_one_channel(800),
            )
        )
        flchcy_ims = ch0_ims[:, None, None, :, :]
        reg_psf = RegPSF.fixture()
        ch_ims = im_proc.fg_estimate(flchcy_ims[:, :, 0], reg_psf, bandpass_kwargs)
        assert ch_ims.shape == (1, 512, 512)
        reg_bal = imops.region_map(ch_ims[0], utils.np_safe_nanmedian, n_divs=5)
        reg_bal = reg_bal / np.nanmax(reg_bal)
        assert 0.95 < reg_bal[2, 2] < 1.05
        assert 0.65 < reg_bal[0, 0] < 0.85

    @zest.retry(2)
    def it_handles_multi_channel():
        flchcy_ims = np.stack(
            (
                _gen_field_two_channel(),
                _gen_field_two_channel(),
                _gen_field_two_channel(),
            )
        )[:, :, None, :, :]

        reg_psf = RegPSF.fixture(n_channels=2)
        ch_ims = im_proc.fg_estimate(flchcy_ims[:, :, 0], reg_psf, bandpass_kwargs)
        assert ch_ims.shape == (2, 512, 512)

        reg_bal = imops.region_map(ch_ims[0], utils.np_safe_nanmedian, n_divs=5)
        reg_bal = reg_bal / np.nanmax(reg_bal)
        assert 0.95 < reg_bal[2, 2] < 1.05
        assert 0.65 < reg_bal[0, 0] < 0.85, f"Failed reg_bal check {reg_bal[0, 0]}"

        reg_bal = imops.region_map(ch_ims[1], utils.np_safe_nanmedian, n_divs=5)
        reg_bal = reg_bal / np.nanmax(reg_bal)
        assert 0.95 < reg_bal[2, 2] <= 1.0
        assert 0.95 < reg_bal[0, 0] <= 1.0

    def it_handles_all_nan_regions():
        ch0_ims = np.stack((_gen_field_one_channel(20), _gen_field_one_channel(20),))
        flchcy_ims = ch0_ims[:, None, None, :, :]
        reg_psf = RegPSF.fixture()
        ch_ims = im_proc.fg_estimate(flchcy_ims[:, :, 0], reg_psf, bandpass_kwargs)
        assert ch_ims.shape == (1, 512, 512)
        reg_bal = imops.region_map(ch_ims[0], utils.np_safe_nanmedian, n_divs=5)
        reg_bal = reg_bal / np.nanmax(reg_bal)
        assert np.any(np.isnan(reg_bal))

    zest()


def zest_fit_peaks_one_im():
    def it_fits():
        reg_psf = RegPSF.fixture(sig_x=1.7, sig_y=1.7)
        locs = np.array([[50.5, 50.5]])
        true_params = (100.0, 1.7, 1.7, 50.5, 50.5, 0.0, 0.0, 100)
        im = imops.gauss2_rho_form(*true_params)
        a = im_proc.fit_peaks_one_im(im, locs, 0, reg_psf)
        assert np.isclose(a[0, 0], 100.0)
        assert np.isclose(a[0, 1], 1.7)
        assert np.isclose(a[0, 2], 1.7)

    zest()
