import numpy as np
from zest import zest
from plumbum import local
from plaster.run.sigproc_v2 import synth
from plaster.tools.test_tools.test_tools import (
    integration_before,
    integration_after,
    run_p,
)
from plaster.run.sigproc_v2 import sigproc_v2_worker as worker
from plaster.run.calib.calib import RegPSF, RegIllum
from plaster.run.sigproc_v2.sigproc_v2_task import SigprocV2Params
from plaster.run.calib.calib import Calib
from plaster.run.sigproc_v2.c_gauss2_fitter.gauss2_fitter import Gauss2Params
from plaster.tools.log.log import debug, prof
from plaster.tools.zlog.zlog import spy


@zest.group("integration")
def zest_sigproc_v2_integration():
    # If test_folder is None it will be set by _before()
    # otherwise you can set it here to reuse a debugging folder
    test_folder = None
    # test_folder = (
    #     "/erisyon/internal/jobs_folder/_integration_tests/it_runs_sigproc_v2/1613159055"
    # )

    dim = (512, 512)
    amps_constant_val = 6000
    n_peaks = 1000
    bg_mean = 100.0
    bg_std = 30.0
    channel_scale_factor = (1.0, 0.5)
    channel_peak_width_scale_factor = (1.0, 0.8)
    psf_width = 1.5

    def _synth_field(n_cycles):
        with synth.Synth(n_channels=2, n_cycles=n_cycles, dim=dim) as s:
            peaks = (
                synth.PeaksModelGaussianCircular(n_peaks=n_peaks)
                .locs_randomize()
                .widths_uniform(psf_width)
                .channel_scale_factor(channel_scale_factor)
                .channel_peak_width_scale_factor(channel_peak_width_scale_factor)
            )
            synth.CameraModel(bg_mean=bg_mean, bg_std=bg_std)
            synth.HaloModel()
            synth.IlluminationQuadraticFalloffModel()

        return s, peaks

    def _synth_save_field(s, fl_i, source_folder):
        chcy_ims = s.render_chcy(0)

        for ch_i in range(chcy_ims.shape[0]):
            for cy_i in range(chcy_ims.shape[1]):
                np.save(
                    str(
                        source_folder
                        / f"area_{fl_i:03d}_cell_000_{ch_i:03d}nm_{cy_i:03d}.npy"
                    ),
                    chcy_ims[ch_i, cy_i],
                )

    def _synth_one_count(source_folder):
        n_fields = 2
        for fl_i in range(n_fields):
            s, peaks = _synth_field(n_cycles=2)
            peaks.amps_constant(val=amps_constant_val)
            _synth_save_field(s, fl_i, source_folder)

    def _synth_dyts(source_folder):
        n_fields = 2
        for fl_i in range(n_fields):
            s, peaks = _synth_field(n_cycles=4)
            peaks.dyt_random_choice(
                [[3, 2, 2, 1], [2, 1, 1, 0], [1, 0, 0, 0]], [0.5, 0.3, 0.2]
            )
            peaks.gain_constant(amps_constant_val)
            _synth_save_field(s, fl_i, source_folder)

    def _it_runs_calib():
        job_folder = test_folder / "calib_synth"

        source_folder = test_folder / "calib_source"
        source_folder.mkdir()

        _synth_one_count(source_folder)

        run_p(
            [
                f"gen",
                f"sigproc_v2_calib",
                f"--mode=illum",
                f"--job={job_folder}",
                f"--instrument_identity=_synth",
                f"--sigproc_source={source_folder}",
                f"--force",
            ]
        )

        run_p(["run", job_folder, "--no_progress"])

    def _it_runs_sigproc_v2():
        job_folder = test_folder / "sigproc"
        calib_file = (
            test_folder
            / "calib_synth/sigproc_v2_calib/plaster_output/sigproc_v2/calib.calib"
        )
        source_folder = test_folder / "sigproc_source"
        source_folder.mkdir()

        _synth_dyts(source_folder)

        run_p(
            [
                f"gen",
                f"sigproc_v2",
                f"--job={job_folder}",
                f"--calibration_file={calib_file}",
                f"--instrument_identity=_synth",
                f"--sigproc_source={source_folder}",
                f"--force",
            ]
        )

        run_p(["run", job_folder, "--no_progress"])

    def _it_runs_sigclass():
        """
        Test that classification can run on sigprov_v2 data
        """

        csv_file = test_folder / "protein.csv"
        job_folder = test_folder / "sigproc_v2_class"
        calib_file = (
            test_folder
            / "calib_synth/sigproc_v2_calib/plaster_output/sigproc_v2/calib.calib"
        )
        source_folder = test_folder / "sigproc_source"
        source_folder.mkdir()

        csv_contents = """
            Name,UniprotAC,Abundance,POI
            ACE,P12821,28,1
            ACE2,Q9BYF1,12,1
            ACTB,P60709,7,1
        """
        csv_contents = "\n".join(csv_contents.split()) + "\n"
        csv_file.write(csv_contents)

        _synth_dyts(source_folder)

        # DEBUGGING HINT: Comment out following to debug report only
        run_p(
            [
                f"gen",
                f"classify_v2",
                f"--sample=KidneyBiomarkers1",
                f"--job={job_folder}",
                f"--protein_csv={csv_file}",
                f"--label_set=C,DE",
                f"--protease=trypsin",
                f"--n_pres=1",
                f"--n_mocks=0",
                f"--n_edmans=3",
                f"--nn_v2",
                f"--rf_v2",
                f"--force",
                f"--field_quality_thresh=0",
                f"--calibration_file={calib_file}",
                f"--instrument_identity=_synth",
                f"--sigproc_source={source_folder}",
            ]
        )

        run_p(["run", job_folder, "--no_progress"])

    def _before():
        nonlocal test_folder
        if test_folder is None:
            test_folder = integration_before()
        test_folder = local.path(test_folder)

    def _after():
        integration_after()

    def it_runs_sigproc_v2():
        # zest randomizes order so the following ensures order...

        # WHEN DEBUGGING:
        #  * Set the test_folder to the folder that failed rather then None
        #    (indicated in the blue trace).
        #  * Comment out the following tests that you don't need to iterate on
        #    (Example, _it_runs_calib and _it_runs_sigproc_v2 has already passed
        #    and you just want to run _it_runs_sigclass over and over until it passes)

        _it_runs_calib()
        _it_runs_sigproc_v2()
        _it_runs_sigclass()

    zest()


@zest.group("integration")
def zest_sigproc_v2_worker_analyze():
    """
    Test the whole sigproc_v2 stack from top to bottom
    """

    def _run(reg_psf, s, extra_params=None):
        if extra_params is None:
            extra_params = {}

        sigproc_v2_params = SigprocV2Params(
            instrument_identity="synth",
            divs=5,
            peak_mea=13,
            calibration_file="",
            mode="analyze",
            **extra_params,
        )

        calib = Calib()
        calib.add_reg_psf(reg_psf, "synth")
        reg_illum = RegIllum(1, reg_psf.im_mea, 5)
        calib.add_reg_illum(reg_illum, "synth")

        ims_import_result = synth.synth_to_ims_import_result(s)
        sigproc_v2_result = worker.sigproc_analyze(
            sigproc_v2_params, ims_import_result, progress=None, calib=calib
        )
        sigproc_v2_result.save()
        return sigproc_v2_result

    def it_returns_exact_distance_no_noise_one_cycle_one_peak():
        """
        Proves that peak finder is picking up the correct locations of sub-pixel positions
        """
        from scipy.spatial.distance import cdist  # Defer slow import

        with synth.Synth(n_channels=1, n_cycles=1, dim=(512, 512)) as s:
            # Change the PSF in the corner to ensure that it is picking up the PSF
            reg_psf = RegPSF.fixture()

            peaks = synth.PeaksModelPSF(reg_psf, n_peaks=1).gain_constant(5000)
            peaks.locs = np.array([[100.3, 72.9]])

            sigproc_v2_result = _run(reg_psf, s)

            dists = cdist(peaks.locs, sigproc_v2_result.locs(), "euclidean")
            closest_iz = np.argmin(dists, axis=1)
            dists = dists[np.arange(dists.shape[0]), closest_iz]
            assert np.all(dists < 0.05)

    def it_returns_exact_sig_from_no_noise_no_collisions_no_bg_subtract():
        """
        Proves that the radiometry is exactly recovering many peaks with
        sub-pixel locations.
        """
        with synth.Synth(n_channels=1, n_cycles=1, dim=(512, 512)) as s:
            reg_psf = RegPSF.fixture()

            (
                synth.PeaksModelPSF(reg_psf, n_peaks=100)
                .locs_grid(pad=50)
                .amps_constant(5000)
                .locs_add_random_subpixel()
            )

            sigproc_v2_result = _run(reg_psf, s, dict(low_inflection=-10.0))

            sig = sigproc_v2_result.sig()[:, 0, 0]
            # np.save("/erisyon/internal/_sig.npy", sig)

            # TODO
            # Somehow with the PSF this shifted slightly from 5000
            # to 4998. For now, I'm accepting this but needs investigation.
            assert np.all(np.abs(sig - 4998) < 0.75)

    def it_returns_exact_sig_from_no_noise_no_collisions_with_bg_subtract():
        """
        Proves that radiometry continues to get consistent answers when
        background subtraction is enabled.

        This is the same as it_returns_exact_sig_from_no_noise_no_collisions_no_bg_subtract()
        but WITH the bg subtraction turned on.

        It is expected that the background subtraction will lower the brightness of
        the peaks but do so consistently.
        """
        with synth.Synth(n_channels=1, n_cycles=1, dim=(512, 512)) as s:
            reg_psf = RegPSF.fixture()

            (
                synth.PeaksModelPSF(reg_psf, n_peaks=100)
                .locs_grid(pad=50)
                .amps_constant(5000)
                .locs_add_random_subpixel()
            )

            sigproc_v2_result = _run(reg_psf, s)

            sig = sigproc_v2_result.sig()[:, 0, 0]
            # Background subtraction is expected to bring down the mean a little bit
            # In default settings it brings it to 4885
            assert np.all(np.abs(sig - 4885) < 4)

    def it_returns_good_signal_no_noise_multi_peak_multi_cycle():
        """
        Prove that the sub-pixel shifting of cycles does not substantially affect the
        quality of the radiometry.
        """
        with synth.Synth(n_channels=1, n_cycles=3, dim=(512, 512)) as s:
            reg_psf = RegPSF.fixture()

            (
                synth.PeaksModelPSF(reg_psf, n_peaks=800)
                .amps_constant(5000)
                .locs_grid(pad=40)
                .locs_add_random_subpixel()
            )

            sigproc_v2_result = _run(reg_psf, s, dict(low_inflection=-10.0))

            sig = sigproc_v2_result.sig()[:, 0, :]

            # There is a small shift up to 4985
            assert np.all(np.abs(sig - 4985) < 8)

    def it_interpolates_regional_PSF_changes():
        """
        Prove that it handles a non-uniform PSF without considering focus
        """
        with synth.Synth(n_channels=1, n_cycles=1, dim=(512, 512)) as s:
            reg_psf = RegPSF.fixture_variable()

            (
                synth.PeaksModelPSF(reg_psf, n_peaks=500)
                .amps_constant(5000)
                .locs_grid()
                .locs_add_random_subpixel()
            )

            sigproc_v2_result = _run(reg_psf, s, dict(low_inflection=-10.0))

            sig = sigproc_v2_result.sig()[:, 0, :]

            # Currently getting mean 4991 which seems a little high but acceptable at moment
            assert np.std(sig) < 20
            assert np.abs(5000.0 - np.mean(sig)) < 10.0

    def it_operates_sanely_with_noise_uniform_psf():
        """A realistic total test that will need to be forgiving of noise, collisions, etc."""
        with synth.Synth(n_channels=1, n_cycles=3, dim=(512, 512)) as s:
            reg_psf = RegPSF.fixture()

            (
                synth.PeaksModelPSF(reg_psf, n_peaks=500)
                .amps_constant(5000)
                .locs_grid(pad=20)
                .locs_add_random_subpixel()
            )
            synth.CameraModel(100, 30)
            synth.HaloModel()

            sigproc_v2_result = _run(reg_psf, s)

            sig = sigproc_v2_result.sig()[:, 0, :]
            assert np.percentile(sig, 20) > 4850.0

    # def it_applies_regional_balance():
    #     """It corrects for non-uniform illumination"""
    #     raise NotImplementedError
    #
    # def it_operates_sanely_with_noise_variable_psf():
    #     """"""
    #     raise NotImplementedError
    #
    # def it_operates_sanely_with_noise_variable_psf_random_locations():
    #     """"""
    #     raise NotImplementedError
    #
    # def it_detects_signal_drops_over_cycles():
    #     """"""
    #     raise NotImplementedError
    #
    # def it_operates_sanely_with_noise_variable_psf_random_locations_non_uniform_lighting():
    #     """"""
    #     raise NotImplementedError
    #
    # def it_measures_aspect_ratio():
    #     """It returns high aspect ratio for collisions under a controlled two-peak system as the peaks approach each other"""
    #     raise NotImplementedError
    #
    # def it_runs_fitter_if_requested():
    #     """It will run the fitter over every peak"""
    #     raise NotImplementedError
    #
    # def it_finds_peaks_as_density_increases():
    #     """Characterize how density affects quality"""
    #     raise NotImplementedError

    zest()


@zest.group("integration")
def zest_sigproc_v2_worker_analyze_multichannel():
    """
    Test the whole sigproc_v2 stack from top to bottom on 2 channels

    1. Calibrate multi-channels separately
    2. Analyze 1 field 2 channels, 3 cycles
    """

    dim = (512, 512)
    amp = 5000.0
    n_peaks = 500  # 1000
    bg_mean = 100.0
    bg_std = 10.0
    channel_scale_factor = (1.0, 0.5)
    channel_peak_width_scale_factor = (1.0, 1.2)
    psf_width = 1.5

    def _synth_field(n_cycles):
        with synth.Synth(n_channels=2, n_cycles=n_cycles, dim=dim) as s:
            peaks = (
                synth.PeaksModelGaussianCircular(n_peaks=n_peaks)
                .locs_randomize()
                .widths_uniform(psf_width)
                .amps_constant(val=amp)
                .channel_scale_factor(channel_scale_factor)
                .channel_peak_width_scale_factor(channel_peak_width_scale_factor)
            )
            synth.CameraModel(bg_mean=bg_mean, bg_std=bg_std)
            synth.HaloModel()
            synth.IlluminationQuadraticFalloffModel()

        return s, peaks

    def _calibrate(s, extra_params=None):
        ims_import_result = synth.synth_to_ims_import_result(s)

        if extra_params is None:
            extra_params = {}

        sigproc_v2_params = SigprocV2Params(
            instrument_identity="_synth", mode="illum_calib", **extra_params,
        )
        sigproc_v2_result = worker.sigproc_instrument_calib(
            sigproc_v2_params, ims_import_result, progress=None
        )
        sigproc_v2_result.save()
        return sigproc_v2_result

    def _analyze(s, extra_params=None):
        ims_import_result = synth.synth_to_ims_import_result(s)

        if extra_params is None:
            extra_params = {}

        sigproc_v2_params = SigprocV2Params(
            self_calib=True, mode="analyze", **extra_params,
        )

        sigproc_v2_result = worker.sigproc_analyze(
            sigproc_v2_params, ims_import_result, progress=None
        )
        sigproc_v2_result.save()
        return sigproc_v2_result

    @zest.retry(2)
    def it_calibrates():
        s, peaks = _synth_field(n_cycles=1)
        sigproc_v2_result = _calibrate(s)

        mean_peak_width_ch0 = np.mean(
            sigproc_v2_result.calib.reg_psf().params[0, :, :, 0]
        )
        mean_peak_width_ch1 = np.mean(
            sigproc_v2_result.calib.reg_psf().params[1, :, :, 0]
        )
        assert 1.4 < mean_peak_width_ch0 < 1.6
        assert 1.7 < mean_peak_width_ch1 < 1.9

        # Note channel 1 is half as bright which means its
        # balance (correction factor) should be 2 so
        # 1 / 2 == 0.5
        ch_bal = sigproc_v2_result.calib.reg_illum().channel_balance()
        assert 0.45 < ch_bal[0] / ch_bal[1] < 0.55

        zest()

    @zest.retry(3)
    def analyzes():
        s, peaks = _synth_field(n_cycles=3)
        sigproc_v2_result = _analyze(s)

        def it_gives_reasonable_sig():
            sig = np.nanmedian(sigproc_v2_result.sig(), axis=(0, 2))
            assert np.all((4500 < sig[0]) & (sig[0] < 5200))
            assert np.all((2000 < sig[1]) & (sig[1] < 2500))

        def it_gives_reasonable_fits():
            fit_a = np.nanmedian(
                sigproc_v2_result.fitmat()[:, :, :, Gauss2Params.AMP], axis=(0, 2)
            )
            fit_x = np.nanmedian(
                sigproc_v2_result.fitmat()[:, :, :, Gauss2Params.SIGMA_X], axis=(0, 2),
            )
            fit_y = np.nanmedian(
                sigproc_v2_result.fitmat()[:, :, :, Gauss2Params.SIGMA_Y], axis=(0, 2),
            )
            fit_r = np.nanmedian(
                sigproc_v2_result.fitmat()[:, :, :, Gauss2Params.RHO], axis=(0, 2)
            )

            assert (4500 < fit_a[0]) & (fit_a[0] < 5200)
            assert (2200 < fit_a[1]) & (fit_a[1] < 2800)
            assert (1.4 < fit_x[0]) & (fit_x[0] < 1.6)
            assert (1.7 < fit_x[1]) & (fit_x[1] < 1.9)
            assert (1.4 < fit_y[0]) & (fit_y[0] < 1.6)
            assert (1.7 < fit_y[1]) & (fit_y[1] < 1.9)
            assert np.all(fit_r < 0.05)

        zest()

    zest()
