import numpy as np
from zest import zest
from scipy.spatial.distance import cdist
from plaster.run.sigproc_v2 import sigproc_v2_worker as worker
from plaster.run.sigproc_v2 import synth
from plaster.run.calib.calib import RegPSF, RegIllum
from plaster.run.sigproc_v2.sigproc_v2_task import SigprocV2Params
from plaster.run.calib.calib import Calib
from plaster.tools.image import imops
from plaster.tools.image.coord import XY
from plaster.run.sigproc_v2.c_gauss2_fitter.gauss2_fitter import Gauss2Params
from plaster.run.calib.calib import approximate_psf
from plaster.run.sigproc_v2 import im_proc
from plaster.tools.log.log import debug, prof
import logging

log = logging.getLogger(__name__)
from plaster.tools.zlog.zlog import spy


def zest_align_image_stack():
    def _synth():
        with synth.Synth(n_channels=1, n_cycles=3, dim=(512, 512)) as s:
            reg_psf = RegPSF.fixture()
            (
                synth.PeaksModelPSF(reg_psf, n_peaks=500)
                .amps_constant(5000)
                .locs_randomize()
            )
            synth.CameraModel(100, 30)
            synth.HaloModel()
            return s.aln_offsets, s.render_chcy()

    def it_handles_clean():
        true_aln_offsets, chcy_ims = _synth()
        pred_aln_offsets = worker._align_image_stack(chcy_ims[0])
        assert np.all(np.abs(true_aln_offsets - pred_aln_offsets) < 0.1)

    @zest.retry(3)
    def it_handles_smudged():
        true_aln_offsets, chcy_ims = _synth()
        smudge_im = np.full((400, 400), 1000)
        imops.accum_inplace(chcy_ims[0, 0], smudge_im, loc=XY(100, 100))
        pred_aln_offsets = worker._align_image_stack(chcy_ims[0])
        if not np.all(np.abs(true_aln_offsets - pred_aln_offsets) < 0.1):
            debug(np.max(np.abs(true_aln_offsets - pred_aln_offsets)))
        assert np.all(np.abs(true_aln_offsets - pred_aln_offsets) < 0.1)

    zest()


def zest_analyze_step_4_align_stack_of_chcy_ims():
    def _synth():
        ch_scale = (1.0, 1.0)
        ch_aln = (0.0, -5.0)
        with synth.Synth(n_channels=2, n_cycles=3, dim=(512, 512)) as s:
            s.channel_aln_factor(ch_aln)
            (
                synth.PeaksModelGaussianCircular(n_peaks=500)
                .uniform_width_and_heights()
                .amps_constant(5000)
                .locs_randomize()
                .channel_scale_factor(ch_scale)
            )
            return s.aln_offsets, s.render_chcy()

    def it_corrects_channels():
        true_aln_offsets, chcy_ims = _synth()
        (
            chcy_offsets,
            offsets_each_channel,
            ch_offsets,
        ) = worker._analyze_step_3_align_chcy_ims(chcy_ims, None)
        aligned_chcy_ims = worker._analyze_step_4_align_stack_of_chcy_ims(
            chcy_ims, chcy_offsets
        )

        # If they misalign correctly then there will be a lot of
        # new peaks that weren't there. We generated 500 so the
        # assert is that there should be less than that on a combined image

        sum_im = np.sum(aligned_chcy_ims[:, 0], axis=0)
        n_peaks = im_proc.pixel_peak_find_one_im(sum_im, approximate_psf()).shape[0]
        assert n_peaks <= 500

    zest()


def zest_analyze_step_5_find_peaks():
    def it_peak_find_very_different_channel_brightnesses_2_channels():
        """
        Assume the images are already aligned.
        Synth two channels with 0.2 overlap
        Ensure that most of the true peaks in each channel are located
        There will be some collisions so it will not be perfect.
        """

        # fmt: off
        dyts = [
            [[1, 1, 1], [0, 0, 0]],  # On in ch 0
            [[0, 0, 0], [1, 1, 1]],  # On in ch 1
            [[1, 1, 1], [1, 1, 1]],  # Shared
        ]
        # fmt: on

        frac_shared = 0.20
        frac_unshared = (1.0 - frac_shared) / 2
        probs = [frac_unshared, frac_unshared, frac_shared]
        n_peaks = 500
        ch_scale = (1.0, 0.2)
        ch_aln = (0.0, 0.0)
        with synth.Synth(n_channels=2, n_cycles=3, dim=(512, 512)) as s:
            s.channel_aln_factor(ch_aln)
            peaks = (
                synth.PeaksModelGaussianCircular(n_peaks=n_peaks)
                .uniform_width_and_heights()
                .locs_randomize()
                .channel_scale_factor(ch_scale)
                .multichannel_dyt_random_choice(dyts, probs)
                .gain_constant(5000)
            )
            synth.CameraModel(0, 5)
            ims = s.render_chcy()

            approx_psf = approximate_psf()
            found_locs = worker._analyze_step_5_find_peaks(ims, approx_psf, 0)

            """
            In each channel, how many of the TRUE exclusive peaks that were
            visible in that channel were found?
            How many of the true shared were found?

            for each of the three groups:
            cdist fro true set to the found and answer within 1.5 pixels.

            peaks.dyt_iz will have 0 (only channel 0), 1 (only ch 1) 2 = both
            """

            def _compare(filt_true_locs):
                dists = cdist(filt_true_locs, found_locs, "euclidean")
                closest_i = np.argmin(dists, axis=1)
                closest_d = dists[np.arange(filt_true_locs.shape[0]), closest_i]
                new_locs_mask = closest_d > 1.5
                return filt_true_locs.shape[0], (~new_locs_mask).sum()

            n_true_ch0, n_found_ch0 = _compare(
                filt_true_locs=peaks.locs[(peaks.dyt_iz == 0) | (peaks.dyt_iz == 2)]
            )
            n_true_ch1, n_found_ch1 = _compare(
                filt_true_locs=peaks.locs[(peaks.dyt_iz == 1) | (peaks.dyt_iz == 2)]
            )
            assert n_found_ch0 / n_true_ch0 > 0.75
            assert n_found_ch1 / n_true_ch1 > 0.75

    def it_peak_find_very_different_channel_brightnesses_4_channels():
        """
        Test that as the number of channels increases that the de-duping
        continues to work
        """

        # fmt: off
        dyts = [
            [[1, 1, 1], [0, 0, 0], [0, 0, 0], [0, 0, 0]],  # On in ch 0
            [[0, 0, 0], [1, 1, 1], [0, 0, 0], [0, 0, 0]],  # On in ch 1
            [[0, 0, 0], [0, 0, 0], [1, 1, 1], [0, 0, 0]],  # On in ch 2
            [[0, 0, 0], [0, 0, 0], [0, 0, 0], [1, 1, 1]],  # On in ch 3
            [[1, 1, 1], [1, 1, 1], [0, 0, 0], [0, 0, 0]],  # Shared 0, 1
            [[0, 0, 0], [1, 1, 1], [1, 1, 1], [0, 0, 0]],  # Shared 1, 2
            [[0, 0, 0], [0, 0, 0], [1, 1, 1], [0, 0, 0]],  # Shared 2, 3
            [[1, 1, 1], [0, 0, 0], [1, 1, 1], [0, 0, 0]],  # Shared 0, 2
            # Good enough...
        ]
        # fmt: on

        probs = np.array([10, 10, 10, 10, 3, 3, 3, 3])
        probs = probs / np.sum(probs)

        n_peaks = 500
        ch_scale = (1.0, 0.1, 0.1, 0.1)
        ch_aln = (0.0, 0.0, 0.0, 0.0)
        with synth.Synth(n_channels=4, n_cycles=3, dim=(512, 512)) as s:
            s.channel_aln_factor(ch_aln)
            peaks = (
                synth.PeaksModelGaussianCircular(n_peaks=n_peaks)
                .uniform_width_and_heights()
                .locs_randomize()
                .channel_scale_factor(ch_scale)
                .multichannel_dyt_random_choice(dyts, probs)
                .gain_constant(5000)
            )
            synth.CameraModel(0, 5)
            ims = s.render_chcy()

            approx_psf = approximate_psf()
            found_locs = worker._analyze_step_5_find_peaks(ims, approx_psf, 0)

            def _compare(filt_true_locs):
                dists = cdist(filt_true_locs, found_locs, "euclidean")
                closest_i = np.argmin(dists, axis=1)
                closest_d = dists[np.arange(filt_true_locs.shape[0]), closest_i]
                new_locs_mask = closest_d > 1.5
                return filt_true_locs.shape[0], (~new_locs_mask).sum()

            n_true_ch0, n_found_ch0 = _compare(
                filt_true_locs=peaks.locs[
                    (peaks.dyt_iz == 0) | (peaks.dyt_iz == 4) | (peaks.dyt_iz == 7)
                ]
            )
            assert n_found_ch0 / n_true_ch0 > 0.80

            n_true_ch1, n_found_ch1 = _compare(
                filt_true_locs=peaks.locs[
                    (peaks.dyt_iz == 1) | (peaks.dyt_iz == 4) | (peaks.dyt_iz == 5)
                ]
            )
            assert n_found_ch1 / n_true_ch1 > 0.80

            n_true_ch2, n_found_ch2 = _compare(
                filt_true_locs=peaks.locs[
                    (peaks.dyt_iz == 2)
                    | (peaks.dyt_iz == 5)
                    | (peaks.dyt_iz == 6)
                    | (peaks.dyt_iz == 7)
                ]
            )
            assert n_found_ch2 / n_true_ch2 > 0.80

            n_true_ch3, n_found_ch3 = _compare(
                filt_true_locs=peaks.locs[(peaks.dyt_iz == 3)]
            )
            assert n_found_ch3 / n_true_ch3 > 0.80

    def it_handles_empty_cycle():
        """
        It doesn't choke if a cycle fails and there's no data
        """
        ims = np.zeros((1, 1, 512, 512))
        approx_psf = approximate_psf()
        found_locs = worker._analyze_step_5_find_peaks(ims, approx_psf, 0)
        assert found_locs.shape == (0, 2)

    zest()
