import re
import json
from plumbum import local
from zest import zest
from plaster.tools.utils.tmp import tmp_folder


def zest_log_integration():
    file_and_line_pat = re.compile(r"^example_module\.py:(\d+)]")

    rc, so, se, lc, jl = (None,) * 5

    def _run(func_name, plaster_zlog_config_path=None):
        with tmp_folder(chdir=True):
            nonlocal rc, so, se, lc, jl
            here_folder = local.path(__file__).dirname
            if plaster_zlog_config_path is None:
                plaster_zlog_config_path = here_folder / "../example_zlog.yaml"
            with local.env(PLASTER_ZLOG_CONFIG_PATH=plaster_zlog_config_path):
                main = here_folder / "./example_main.py"
                rc, so, se = local.python.run(("-u", main, func_name,), retcode=None)
                log_file = local.path("plaster_dev.log")
                lc, jl = None, None
                if log_file.exists():
                    lc = log_file.read()
                    jl = [json.loads(i) for i in lc.split("\n") if i != ""]

    def _dump():
        print(f"{rc=}")
        print(f"{so=}")
        print(f"{se=}")
        print(f"{lc=}")
        print(f"{jl=}")

    def startup_config():
        def dev_mode():
            _run("normal_traces")

            def it_captures_root_logger():
                assert len(jl) == 1
                assert (
                    jl[0]["message"] == "example_module.py wrote to root logger.debug"
                )

            def it_does_not_capture_stdio_to_log_file():
                assert "wrote to stdout" not in lc
                assert "wrote to stderr" not in lc

            def it_echos_log_to_stdout_human_readable():
                assert "example_module.py wrote to root logger.debug" in so

            def it_does_not_intercept_to_stdio():
                assert "wrote to stdout" in so
                assert "wrote to stderr" in se

            zest()

        def it_allows_env_set_for_the_yaml_file():
            with tmp_folder(chdir=True, remove=False) as t:
                lines = [
                    "version: 1",
                    "formatters:",
                    "  json:",
                    "    class: plaster.tools.zlog.zlog.TypeAwareJsonFormatter",
                    "    format: '%(asctime)s %(levelname)s %(message)s %(filename)s %(lineno)d'",
                    "handlers:",
                    "  console:",
                    "    class: logging.StreamHandler",
                    "    stream: ext://sys.stdout",
                    "    formatter: json",
                    "    level: DEBUG",
                    "loggers:",
                    "  plaster:",
                    "    level: DEBUG",
                    "    handlers: [ console ]",
                    "    propagate: 0",
                ]
                with open(t / "logger.yaml", "w") as f:
                    f.write("\n".join(lines))
                _run("normal_traces", plaster_zlog_config_path=t / "logger.yaml")
                lines = so.split("\n")
                s = json.loads(lines[0])
                assert "wrote to root logger.debug" in s["message"]

        def it_gives_a_meaningful_error_when_the_yaml_file_isnt_found():
            _run("normal_traces", plaster_zlog_config_path="does_not_exist")
            assert "FATAL LOG STARTUP" in se
            assert "was set to 'does_not_exist'" in se

        zest()

    def spy():
        def it_emits_formatted_args_and_vals_to_human_readable():
            _run("spies_int")
            assert "a:int = 1\n" in so and "b:int = 2\n" in so

        def it_emits_formatted_args_and_vals_to_json():
            _run("spies_int")
            assert jl[0]["message"] == "a:int = 1"
            assert jl[1]["message"] == "b:int = 2"

        def it_emits_file_and_line_to_human_readable():
            _run("spies_int")
            mat = file_and_line_pat.match(so)
            assert mat

        def it_shows_types_for_python_vars():
            _run("spies_mixed")
            assert "a_var:str = but" in jl[0]["message"]
            assert "constant:str = why?!" in jl[1]["message"]
            assert "constant:int = 345" in jl[2]["message"]
            assert "constant:tuple = (1, 2, 3)" in jl[3]["message"]
            assert "hello:str = world" in jl[4]["message"]

        def it_shows_types_for_numpy():
            _run("spies_array")
            assert "a:ndarray[2]{int64} = [1 2]" in jl[0]["message"]
            assert "a:ndarray[2]{int64} = [1 2]" in so

        zest()

    def human_readable_includes_file_and_lineno():
        _run("normal_traces")
        mat = file_and_line_pat.match(so)
        assert mat

    def exceptions():
        _run("raises_an_exception")

        def it_has_human_readable():
            assert "was raised with message:" in so
            assert "test exception" in so

        def it_has_machine_readable():
            assert "in raises_an_exception" in jl[0]["exc_info"]
            assert "test exception" in jl[0]["exc_info"]

        zest()

    def it_handles_headless_pudb_imports():
        _run("run_pudb")
        assert "You left a pudb breakpoint in" in jl[0]["message"]
        assert "It appears to be at:   File" in jl[0]["message"]

    def it_can_add_new_records_from_env():
        with local.env(PLASTER_LOG_FIELD_FOO=1):
            _run("normal_traces")
            assert jl[0]["foo"] == "1"

    def it_can_add_new_records_at_runtime():
        _run("add_new_field")
        assert jl[0]["bar"] == "123"

    # def it_handles_multithread():
    #     raise NotImplementedError
    #
    # def it_handles_multiprocess():
    #     raise NotImplementedError

    # To go in to integration tests
    # def jupyter():
    #     def it_logs_from_jupyter_explicitly():
    #         raise NotImplementedError
    #
    #     def it_traps_lib_errors():
    #         raise NotImplementedError
    #
    #     def it_does_not_trap_in_interactive_mode():
    #         # hard to test in an integration test
    #         raise NotImplementedError
    #

    # def startup_log_catches():
    # This will go in to the e2e tests
    # Test the ways that a program can fail to conform to the
    # python logger and make sure those are captured.
    # This is the responsibility of the docker entry point scripts

    # def it_captures_stray_print_commands():
    #     raise NotImplementedError
    #
    # def it_captures_OOM_errors():
    #     raise NotImplementedError
    #
    # def it_captures_syntax_errors():
    #     raise NotImplementedError
    #

    # To go into the stand c helpers tests
    # def c():
    #     def it_asserts():
    #         raise NotImplementedError
    #
    #     def it_traces():
    #         raise NotImplementedError
    #
    #     def it_segfaults():
    #         raise NotImplementedError
    #

    zest()
