"""
Usage patterns:

Initialization at main entrypoint:

    # Ensure that either:
    #   * logger.yaml is found parallel to the entrypoint
    #   * --OR-- that

    if __name__ == "__main__":
        from plaster.tools.zlog import zlog
        with zlog.app_start():
            do_stuff()

Initialization for each module follows Python standard:

    from logging import getLogger
    log = getLogger(__name__)
    log.debug("blah")

    # But you can also add spy.
    from plaster.tools.zlog.zlog import spy
    from logging import getLogger
    log = getLogger(__name__)
    foo = 1
    spy(foo)


"""

import os
import re
import sys
import numpy as np
from contextlib import contextmanager
from plumbum import local, colors
import traceback
import logging
import logging.config
import yaml
import inspect
import ast
from pythonjsonlogger.jsonlogger import JsonFormatter


# Ths tb_pat regular expression is used to break up traceback for colorization
tb_pat = re.compile(r"^.*File \"([^\"]+)\", line (\d+), in (.*)")


def human_readable_type_and_value(arg):
    """
    Examine the type of arg and emit it in friendly ways.
    The set of types that are given special consideration is
    likely to grow. At moment, it only includes numpy but will
    likely containe special cases for pandas or other library types
    that we need special type info.
    """

    type_str = str(type(arg).__name__)
    val_str = str(arg)

    if isinstance(arg, np.ndarray):
        dims = ",".join([f"{dim}" for dim in arg.shape])
        type_str = f"ndarray[{dims}]{{{arg.dtype}}}"
        val_str = str(arg)

    return type_str, val_str


class ColorfulFormatter(logging.Formatter):
    """
    A formatter that adds color, especially to exceptions
    but also to any receod that contains spy fields (see def spy)
    """

    def _colorful_exception(self, error=None, formatted=None, msg=""):
        s = ""

        def _traceback_match_filename(line):
            is_libs = False
            m = tb_pat.match(line)
            if m:
                file = m.group(1)
                lineno = m.group(2)
                context = m.group(3)
                real_path = os.path.realpath(file)
                relative_path = os.path.relpath(real_path)

                root = os.environ.get("ERISYON_ROOT")
                if root is not None:
                    is_libs = True
                    if real_path.startswith(root):
                        relative_path = re.sub(r".*/" + root, "./", real_path)
                        is_libs = False

                if real_path == __file__:
                    # Any functions in the log helper are "libs"
                    is_libs = True

                # Treat these long but commonly occurring path differently
                if "/site-packages/" in relative_path:
                    relative_path = re.sub(r".*/site-packages/", ".../", relative_path)
                if "/dist-packages/" in relative_path:
                    relative_path = re.sub(r".*/dist-packages/", ".../", relative_path)

                leading, basename = os.path.split(relative_path)
                # if leading and len(leading) > 0:
                #     leading = f"{'./' if leading[0] != '.' else ''}{leading}"
                return leading, basename, lineno, context, is_libs

            return None

        s += colors.red & colors.bold | error.__class__.__name__
        s += "('"
        error_message = str(error).strip()
        if error_message != "":
            s += colors.red | error_message
        s += "')\n"
        s += f"  was raised with message: "
        s += colors.red | f"'{msg}'\n"

        if formatted is None:
            formatted = traceback.format_exception(
                etype=type(error), value=error, tb=error.__traceback__
            )

        lines = []
        for line in formatted:
            lines += [sub_line for sub_line in line.strip().split("\n")]

        is_libs = False
        for line in lines[1:-1]:
            split_line = _traceback_match_filename(line)
            if split_line is None:
                # The is_libs flag is set every other line and remains
                # set for the followig line which shows the context
                s += (colors.DarkGray if is_libs else colors.white) | line + "\n"
            else:
                leading, basename, lineno, context, is_libs = split_line
                s += "  "
                if is_libs:
                    s += colors.DarkGray | leading + "/"
                    s += colors.DarkGray | basename + ":"
                    s += colors.DarkGray | lineno + " in function "
                    s += colors.DarkGray | context + "\n"
                else:
                    s += colors.yellow | leading + "/"
                    s += colors.yellow & colors.bold | basename
                    s += ":"
                    s += colors.yellow & colors.bold | lineno
                    s += " in function "
                    s += colors.magenta & colors.bold | context + "\n"

        return s

    def format(self, record: logging.LogRecord) -> str:
        type_str, val_str = human_readable_type_and_value(record.msg)

        if hasattr(record, "spy_variable_name"):
            # This record contains spy variable fields which are
            # colorized differntly. See def spy() to see how these fields
            # go into this record.
            msg = (
                (colors.green | f"{record.spy_variable_name}")
                + (colors.DarkGray | ":")
                + (colors.yellow | type_str)
                + (colors.DarkGray | " = ")
                + (colors.white | val_str)
            )
        else:
            msg = colors.green | str(record.msg)

        if hasattr(record, "exc_info") and record.exc_info is not None:
            # The case of an expcetion use the fancy colorful exception formatter
            e = record.exc_info[1]
            formatted = traceback.format_exception(
                etype=type(e), value=e, tb=e.__traceback__
            )
            msg = self._colorful_exception(error=e, formatted=formatted, msg=record.msg)
        else:
            msg = (colors.blue | f"{record.filename}:{record.lineno}] ") + msg

        return msg


class TypeAwareJsonFormatter(JsonFormatter):
    """
    Adds human readable type information to json log messages when the spy()
    function adds the "spy_variable_name" to the log record
    """

    def format(self, record: logging.LogRecord) -> str:
        if hasattr(record, "spy_variable_name"):
            verbose_type_str, val_str = human_readable_type_and_value(record.msg)
            record.msg = f"{record.spy_variable_name}:{verbose_type_str} = {val_str}"
        return super().format(record)


# def findCaller(self, stack_info=False, stacklevel=1):
#     rv = super().findCaller(stack_info, stacklevel=2)
#     if rv[0] == __file__ and rv[2] == "spy":
#         # Found the spy meaning we need to go up one more level
#         rv = super().findCaller(stack_info, stacklevel=3)
#     return rv


def spy(*args, log=None, **kwargs):
    """
    Example usage:

        foo = 1
        log.spy(foo)
        # Emits: file.py:10] foo:int = 1
        # Or to a JSON file it would emit "foo:int = 1" to the message record

    Example multi-parameter usage:

        foo = 1
        bar = "WTF?"
        log.spy(foo, bar)
        # Emits: file.py:10] foo:int = 1, bar:str = "WTF?"

    Example of an expression.
    If you need a the value of temporary expression but don't
    want to make a temporary variable, just assign as a kwarg
    in the spy() call.

        foo = np.array([...])
        log.spy(sqrt_foo=np.sqrt(foo))
        # Emits: file.py:10] foo:ndarray[...] = [...]

    By default writes to the root logger or optionally pass
    a logger instance in to this using the log= kwargs

    Example:
        log = getLogger(__name__)
        spy(foo, log=log)

    """
    if log is None:
        log = logging.getLogger("")  # Use root logger

    def _find_caller_node(root_node, func_name, last_lineno):
        # init search state. Based on examples here:
        # https://stackoverflow.com/questions/28244921/how-can-i-get-the-calling-expression-of-a-function-in-python
        found_node = None
        lineno = 0

        def _ast_walker(parent):
            nonlocal found_node
            nonlocal lineno
            for child in ast.iter_child_nodes(parent):
                if hasattr(child, "lineno"):
                    lineno = child.lineno

                if lineno > last_lineno:
                    # break if we passed the last line
                    break

                if (
                    isinstance(child, ast.Name)
                    and isinstance(parent, ast.Call)
                    and child.id == func_name
                ):
                    found_node = parent
                    break

                # RECURSE walk through children nodes, if any
                _ast_walker(child)

        # dig recursively to find caller's node
        _ast_walker(root_node)
        return found_node

    # FIND the caller
    frame = inspect.currentframe()
    back_frame = frame.f_back

    # get the source code of caller's module. See notes in above stackoverflow citation
    with open(back_frame.f_code.co_filename, "r") as f:
        source = f.read()

    ast_root = ast.parse(source, back_frame.f_code.co_filename)

    # find caller's ast node so that we can extract the arg name
    caller_node = _find_caller_node(ast_root, "spy", back_frame.f_lineno)
    if caller_node:

        def log_key_val(key, val):
            log.debug(val, extra=dict(spy_variable_name=key), stacklevel=3)

        if hasattr(caller_node, "args"):
            for key, val in zip(caller_node.args, args):
                if hasattr(key, "id"):
                    log_key_val(key.id, val)
                else:
                    log_key_val("constant", val)

        if hasattr(caller_node, "keywords"):
            for key, val in zip(caller_node.keywords, kwargs.values()):
                if hasattr(key, "arg"):
                    log_key_val(key.arg, val)
                else:
                    log_key_val("unknown kwarg", val)


def _could_not_find_config_file_error_message(log_config_path):
    env_set = local.env.get("PLASTER_ZLOG_CONFIG_PATH")
    msg = (
        f"\n======================== FATAL LOG STARTUP ERROR: ========================\n"
        f"Plaster's zlog system was not able to find its configuration file.\n"
        f"It first looked at the path specified in the environment variable: 'PLASTER_ZLOG_CONFIG_PATH' "
    )
    if env_set:
        msg += f"""which was set to '{local.env.get("PLASTER_ZLOG_CONFIG_PATH")}'.\n"""
    else:
        msg += (
            "which was not set.\n"
            f"So it looked for '{log_config_path}'\n"
            f"by using the folder of the code that called app_start()\n"
            f"which should be in "
            f"""the 'if __name__ == "__main__":' block."""
        )
    msg += f"\n==========================================================================\n"
    return msg


extra_log_fields = {}


@contextmanager
def add_log_fields(**kwargs):
    global extra_log_fields
    old_extra_log_fields = extra_log_fields.copy()
    extra_log_fields.update(kwargs)
    try:
        yield
    finally:
        extra_log_fields = old_extra_log_fields


config_dict = None


@contextmanager
def app_start():
    """
    Call this function the first thing from the `if __name__ == "__main__":` block

    Example:

        if __name__ == "__main__":
            from plaster.tools.zlog import zlog
            with zlog.app_start():
                start_your_app()

    This is implemented as a contextmanager so that we can
    trap exceptions and display them through the standard
    configured logging system which means that when you include
    a ColorfulFormatter formatter in your config that all
    exceptions that bubble to the top will be displayed
    with colorful print out.

    It also adds special exception handling for the common
    case where a pudb.set_trace() gets left in a deploy to
    a headless terminal and will print a more useful log
    than the hard-to-interpret exception about TypeError
    that occurs when pudb can't get a handle to stdin.

    This also adds fields to log messages that are encoded in the
    environmant variables that start with PLASTER_LOG_FIELD_

    Example:
        PLASTER_LOG_FIELD_DEED=abc p run ...

        This will cause all log messages from that p run
        to include a field "deed":"abc" in their JSON output

    Note that you can also add these fields at runtime

    with add_log_fields(foo=1):
        # During this foo = 1 will be added to all logs
        # This is especially hand for debugging parallel
        # run where a large number of similar workers
        # are running and you need to see context

    """

    # FIND the YAML FILE based o a search order
    #  1. ENV var?
    #  2. Adjacent to the source of the module that called this

    # In dev mode, set PLASTER_ZLOG_CONFIG_PATH which will be set in h dev
    log_config_path = local.env.get("PLASTER_ZLOG_CONFIG_PATH")

    if not log_config_path:
        caller = traceback.extract_stack()[0]
        main_entrypoint_folder = local.path(caller.filename).dirname

        # Search in the folder of the caller of this function
        log_config_path = main_entrypoint_folder / "zlog.yaml"

    global config_dict

    log_config_path = local.path(log_config_path)
    if log_config_path.exists():
        config_dict = yaml.safe_load(log_config_path.read())
    else:
        msg = _could_not_find_config_file_error_message(log_config_path)
        print(msg, file=sys.stderr)
        # This is a sane default but should be set with a yaml file
        # on a per-service level.
        config_dict = {
            "version": 1,
            "formatters": {
                "json": {
                    "class": "plaster.tools.zlog.zlog.TypeAwareJsonFormatter",
                    "format": "%(asctime)s %(levelname)s %(message)s %(filename)s %(lineno)d %(module)s %(name)s %(process)d",
                },
                "colorful": {"class": "plaster.tools.zlog.zlog.ColorfulFormatter"},
            },
            "handlers": {
                "console": {
                    "class": "logging.StreamHandler",
                    "stream": "ext://sys.stdout",
                    "formatter": "colorful",
                    "level": "DEBUG",
                },
                "file": {
                    "class": "logging.FileHandler",
                    "filename": "internal.log",
                    "formatter": "json",
                    "level": "DEBUG",
                },
            },
            "loggers": {
                "": {
                    "level": "DEBUG",
                    "handlers": ["console", "file"],
                    "propagate": 0,
                },
                "plaster": {
                    "level": "DEBUG",
                    "handlers": ["console", "file"],
                    "propagate": 0,
                },
                "matplotlib": {
                    "level": "ERROR",
                    "handlers": ["console", "file"],
                    "propagate": 0,
                },
                "botocore": {
                    "level": "ERROR",
                    "handlers": ["console", "file"],
                    "propagate": 0,
                },
                "urllib3": {
                    "level": "ERROR",
                    "handlers": ["console", "file"],
                    "propagate": 0,
                },
            },
        }
    logging.config.dictConfig(config_dict)

    old_factory = logging.getLogRecordFactory()

    env_pat = re.compile(r"^PLASTER_LOG_FIELD_(.+)")
    for key, val in local.env.items():
        m = env_pat.match(key)
        if m:
            extra_log_fields[m.group(1).lower()] = val

    def plaster_specific_field_factory(*args, **kwargs):
        record = old_factory(*args, **kwargs)
        for key, val in extra_log_fields.items():
            setattr(record, key, val)
        return record

    logging.setLogRecordFactory(plaster_specific_field_factory)

    try:
        yield
    except TypeError as e:
        # Handle the common case where a pudbn set_trace()
        # got left in a headless execution and replace the
        # hard-to-interpret expcetion as a more helpful error message.
        log = logging.getLogger("")  # Use root logger for all top level exceptions
        formatted = traceback.format_exception(
            etype=type(e), value=e, tb=e.__traceback__
        )
        for line in formatted:
            if "pudb/debugger.py" in line:
                found = ""

                for i, line in enumerate(formatted):
                    if "pudb.set_trace()" in line:
                        found = f"It appears to be at: {formatted[i-1]}"
                        break

                log.error(f"You left a pudb breakpoint in!. {found}")
                break
        else:
            raise
    except Exception as e:
        log = logging.getLogger("")  # Use root logger for all top level exceptions
        log.exception(e)
